static const std::string READ_VAL_Common =  R"(

    template<typename CharP>
    void SkipSpace(CharP& input)
    {
        for(;;)
            switch(*input)
            {
                case ' ':  case '\t':
                case '\v':
                case '\r': case '\n':
                    ++input;
                    break;
                default:
                    return;
            }
    }

    template<typename T, bool TrapErrors>
    void B_ReadGeneric(T& target, const char* input)
    {
        constexpr bool FloatType    = B_IsFloatType<T>::value;
        //constexpr bool UnsignedType = B_IsUnsignedType<T>::value;
        constexpr auto minvalue = B_MinValue<T>::value;
        constexpr auto maxvalue = B_MaxValue<T>::value;

        long double         longdouble_result;
        unsigned long long  longlong_result;
        bool negative           = false;
        bool definitely_integer = false;
        unsigned integer_base   = 10;

        SkipSpace(input);

        if(*input == '-') { negative = true; ++input; }
        if(*input == '+') ++input;

        if(*input == '&' && (input[1] == 'H' || input[1] == 'h'))
            { integer_base = 16; definitely_integer = true; input += 2; }
        else if(*input == '&' && (input[1] == 'O' || input[1] == 'o'))
            { integer_base =  8; definitely_integer = true; input += 2; }
        else if(*input == '&' && (input[1] == 'B' || input[1] == 'b'))
            { integer_base =  2; definitely_integer = true; input += 2; }
        else if(*input == '0' && (input[1] == 'X' || input[1] == 'x'))
            { integer_base = 16; definitely_integer = true; input += 2; }
        else if(*input == '0' && (input[1] == 'B' || input[1] == 'b'))
            { integer_base =  2; definitely_integer = true; input += 2; }

        if(unlikely(!*input)) { err: if(TrapErrors) B_ERROR(2); return; } // SYNTAX ERROR

        char*end = 0;
        longlong_result = std::strtol(input, &end, integer_base);

        if(!definitely_integer && (*end == '.' || *end == 'e' || *end == 'E'))
        {
            longdouble_result = std::strtold(input, &end);
            definitely_integer = false;
            if(*end == '!' || *end == '#') ++end;
        }
        else
        {
            longdouble_result = 0.l;
            switch(*end)
            {
                case '%':
                case '&':
                    ++end;
                    definitely_integer = true;
                    break;
                case '!':
                case '#':
                    if(definitely_integer) break;
                    ++end;
                    longdouble_result =  longlong_result;
                    definitely_integer = false;
                    break;
                default:
                    definitely_integer = true;
            }
        }

        SkipSpace(end);
        if(*end && TrapErrors) goto err;

        switch(negative*2 + definitely_integer)
        {
            case 0: // not negative, not integer
                target = longdouble_result;
                if(TrapErrors && unlikely(!FloatType && longdouble_result > maxvalue)) B_ERROR(6); // overflow
                break;
            case 1: // not negative, integer
                target = longlong_result;
                if(TrapErrors && unlikely(!FloatType && longlong_result > maxvalue)) B_ERROR(6); // overflow
                break;
            case 2: // negative, not integer
                target = -longdouble_result;
                if(TrapErrors && unlikely(!FloatType && longdouble_result > -minvalue)) B_ERROR(6); // overflow
                break;
            case 3: // negative, integer
                target = -(long long)longlong_result;
                if(TrapErrors && unlikely(!FloatType && longlong_result > -minvalue)) B_ERROR(6); // overflow
                break;
        }
    }
)";

static const std::string TIMES_DATES_Common = R"(
}
#include <sys/time.h>
namespace
{
    void TrySetTime(const std::string& str)
    {
        int month=0,
              day=0,
             year=0,
             hour=0,
           minute=0,
           second=0;
        int n = std::sscanf(str.c_str(),
            "%d%*[-/]%d%*[-/]%d %d:%d:%d",
            &month,
            &day, 
            &year,
            &hour,  
            &minute, 
            &second);
        if(year >= 0 && year < 100) year += 1900;

        if(n < 3) { Error: B_ERROR(5); } // Illegal function call 

        struct tm tm;
        tm.tm_sec  = second;
        tm.tm_min  = minute;
        tm.tm_hour = hour;
        tm.tm_mday = day;
        tm.tm_mon  = month - 1;   
        tm.tm_year = year  - 1900;
        auto t = std::mktime(&tm); if(t == (decltype(t))-1) goto Error;
        if(std::stime(&t) < 0) B_ERROR_Errno();
    }
)";

static const std::string Func_TIMES_Common = R"(
}
#include <sys/time.h>
namespace
{
    std::string B_FuncTIME_S()
    {
        // hh:mm:ss
        std::time_t t = std::time(0);
        char Buf[512];
        std::strftime(Buf, 512, "%H:%M:%S", std::localtime(&t));
        return Buf;
    }
)";
static const std::string Func_DATES_Common = R"(
}
#include <sys/time.h>
namespace
{
    std::string B_FuncDATE_S()
    {
        // mm-dd-yyyy
        std::time_t t = std::time(0);
        char Buf[512];
        std::strftime(Buf, 512, "%m-%d-%Y", std::localtime(&t));
        return Buf;
    }
)";



static const std::string Func_USING_Common = R"(

/*
                     Characters that format a numeric expression
 #    Digit position.                    | -     Placed after digit, prints
 .    Decimal point position.            |       trailing sign for negative
 ,    Placed left of the decimal point,  |       numbers.
      prints a comma every third digit.  | $$    Prints leading $.
 +    Position of number sign.           | **    Fills leading spaces with *.
^^^^  Prints in exponential format.      | **$   Combines ** and $$.

                     Characters used to format a string expression
 &    Prints entire string.              | \ \   Prints first n characters,
 !    Prints only the first character    |       where n is the number of
      of the string.                     |       blanks between slashes + 2.

                     Characters used to print literal characters
 _    Prints the following formatting    |       Any character not in this
      character as a literal.            |       table is printed as a
                                         |       literal.
NOTES:
    $$ allocates two digit positions just like ## does
              except in scientific notation
    -  terminates the digit sequence (also counts as a character of sequence length)
       Note: If number is not negative, counts as a space
    .  also allocates one digit position
    ,  so does this. Note: Affects only once; position irrelvant.
    +  also allocates one digit position:
        Note: If comes ahead the sequence, indicates that
              sign is to be placed on left side of sequence
              If comes in middle of sequence, terminates the
              digit sequence (still allocates one slot).

    Example on using ^^^^:   (Note: D for doubles, E for others)

     |+#.##^^^^  for 123456 prints:
     |+1.23D+05
     |
     |+##.##^^^^  for 123456 prints:
     |+12.35D+04
     |
     |###.##^^^^^  for 123456 prints:
     | 12.35D+004
     |
     |###.##^^^^^+  for 123456 prints:
     |123.46D+003+
     |
     |$$#.##^^^^   for 123456 prints:
     | $1.23D+05
     |
     |**#.##^^^^   for 123456 prints:
     |*12.34D+04
     |
     |**#.##^^^^  for -123456 prints:
     |-12.34D+04
     |
     |**$.##^^^^   for 123456 prints:
     |*$1.23D+05
     |
     |**$.##^^^^  for -123456 prints:
     |-$1.23D+05
     |
     |#.##^^^^   for 123456 prints:
     |0.12D+06
     |
     |#.##^^^^  for -123456 prints:
     |-.12D+06

*/

class UsingFormatter
{
private:
    struct FormatItem
    {
        enum SignType
        {
            NormalSignLeft,    // number slot or -
            AlwaysSignLeft,    // + or -
            NegativeSignRight, // - or space (fill_char does not affect)
            AlwaysSignRight    // + or -
        };

        enum { Literal, String, Numeric } type;
        // For literals
        std::string literal;
        // For strings
        int         length; // -1=entire string
        // For numeric
        union
        {
            struct
            {
                /* Note: These fields are so constructed that 0 is a sensible
                 *       default for each field. */
                unsigned    num_leading_chars       : 8;
                unsigned    num_trailing_chars      : 8;
                unsigned    scientific              : 2; // 0, 2 or 3
                bool        fill_stars              : 1; // 0=spaces, 1=stars
                bool        decimal_point           : 1;
                bool        comma_every_third_digit : 1;
                bool        leading_dollar          : 1; // note: prints $ instead of last space
                SignType    sign_type               : 2;
                                            //       note: goes _after_ SignLeft
            };
            uint_least32_t numdata;
        };
    };
    std::vector<FormatItem> format;
    size_t pos;
public:
    explicit UsingFormatter(const std::string& f) : format(), pos(0) { CreateFormat(f); }
    ~UsingFormatter() {}

public:
    template<typename T, typename... Rest>
    inline PrintStreamBase& DoParam(PrintStreamBase& out, const T& value, Rest&&... rest)
    {
        return DoParam(Do1param(out, value), std::forward<Rest>(rest)...);
    }

    inline PrintStreamBase& DoParam(PrintStreamBase& out)
    {
        return out;
    }

private:
    /* Note: Do1param() can execute B_ERROR(13) (Type mismatch) */
    PrintStreamBase& Do1param(PrintStreamBase& out, BAS_SGN v)
    {
        return DoParam(out, (BAS_DBL)v, 'E');
    }
    PrintStreamBase& Do1param(PrintStreamBase& out, BAS_DBL v)
    {
        return DoParam(out, v, 'D');
    }
    PrintStreamBase& Do1param(PrintStreamBase& out, const std::string& v)
    {
        FlushLiterals(out, true);

        const FormatItem& format = GetFormat();
        if(format.type != FormatItem::String) B_ERROR(13); // type mismatch

        if(format.length == -1)
            out << v;
        else
        {
            size_t n_eat    = format.length;
            size_t n_spaces = 0;

            if(n_eat > v.size())
            {
                n_spaces = n_eat - v.size();
                n_eat    = v.size();
            }

            if(n_eat    > 0) out << v.substr(0, n_eat);
            if(n_spaces > 0) out << std::string(n_spaces, ' ');
        }

        FlushLiterals(out, false);

        return out;
    }

    PrintStreamBase& Do1param(PrintStreamBase& out, BAS_LNG v)
    {
        // 64-bit IEEE double has enough range to cover a 32-bit signed integer,
        // so to save effort and code, we handle it as a double.
        return DoParam(out, (BAS_DBL)v, 'E');
    }
    PrintStreamBase& Do1param(PrintStreamBase& out, BAS_ULONG v)
    {
        // FIXME: Is this safe?
        return DoParam(out, (BAS_DBL)v, 'E');
    }

    inline PrintStreamBase& Do1param(PrintStreamBase& out, BAS_INT v)   { return Do1param(out, BAS_LNG(v)); }
    inline PrintStreamBase& Do1param(PrintStreamBase& out, BAS_BYTE v)  { return Do1param(out, BAS_LNG(v)); }
    inline PrintStreamBase& Do1param(PrintStreamBase& out, BAS_UBYTE v) { return Do1param(out, BAS_LNG(v)); }
    inline PrintStreamBase& Do1param(PrintStreamBase& out, BAS_UINT v)  { return Do1param(out, BAS_LNG(v)); }

    inline PrintStreamBase& Do1param(PrintStreamBase& out, const char* v) { return Do1param(out, std::string(v)); }
    inline PrintStreamBase& Do1param(PrintStreamBase& out, const char v)  { return Do1param(out, std::string(1,v)); }

    PrintStreamBase& Do1param(PrintStreamBase& out, const BAS_DBL v, char ExpChar)
    {
        FlushLiterals(out, true);

        const FormatItem& format = GetFormat();
        if(format.type != FormatItem::Numeric) B_ERROR(13); // type mismatch

        if(format.scientific)
        {
            //FIXME
            // Note: QB ignores comma in scientific mode.

            int leading_decimals = format.num_leading_chars;
            if(format.sign_type == FormatItem::AlwaysSignLeft
            || format.sign_type == FormatItem::NormalSignLeft)
            {
                --leading_decimals;
            }
            if(format.leading_dollar)
            {
                --leading_decimals;
            }

            // L = 10's exponent to divide v with,
            // so that result is 0.0 <= v < 1.0
            int L = (int) ((1 + 1e-8) + log10( v==0.0 ? 1.0 : std::fabs(v) ));

            L -= leading_decimals;

            double value = v / pow(10.0, L);

            FormatItem tmpformat = format;
            tmpformat.scientific = false;
            if(format.sign_type == FormatItem::NegativeSignRight
            || format.sign_type == FormatItem::AlwaysSignRight)
            {
                tmpformat.sign_type = FormatItem::NormalSignLeft;
                value = std::fabs(value);
            }

            OutputNonscientific(out, value, tmpformat);

            char expo[512];
            sprintf(expo, "%c%+0*d", ExpChar, 1 + format.scientific, L);

            out << expo;

            if(format.sign_type == FormatItem::NegativeSignRight) out << (v < 0 ? '-' : ' ');
            if(format.sign_type == FormatItem::AlwaysSignRight) out << (v < 0 ? '-' : '+');
        }
        else
        {
            OutputNonscientific(out, v, format);
        }

        FlushLiterals(out, false);

        return out;
    }

    const FormatItem& GetFormat()
    {
        if(format.empty())
        {
            B_ERROR(5); // illegal function call
        }

        const FormatItem& result = format[pos];
        if(++pos == format.size()) pos = 0;
        return result;
    }

    void FlushLiterals(PrintStreamBase& out, bool allow_wrap)
    {
        if(format.empty())
        {
        Error:
            B_ERROR(5); // illegal function call
        }

        if(!allow_wrap && pos == 0) return;

        size_t origpos = pos;
        while(format[pos].type == FormatItem::Literal)
        {
            out << format[pos].literal;
            if(++pos == format.size()) pos = 0;
            if(pos == origpos) goto Error; // full loop without finding a parameter

            if(pos == 0 && !allow_wrap) break;
        }
    }

    void CreateFormat(const std::string& f)
    {
        if(false)
        {
        IllCall:
            format.clear(); // this will do IllCall at runtime
            return;
        }

        std::string literal;
        const char* const s = f.c_str();
        size_t spos = 0, slen = f.size();

        #define FlushLiteral() do { \
            if(!literal.empty()) \
            { \
                FormatItem tmp; \
                tmp.type    = FormatItem::Literal; \
                tmp.literal = literal; \
                format.push_back(tmp); \
                literal.clear(); \
            } \
        } while(0)

        while(spos < slen)
        {
            #define remain (slen-spos)
            #define c      (s[spos])

            if(c == '_' && remain >= 2)
            {
                literal += s[spos+1];
                spos += 2; continue;
            }
            if(c == '&')
            {
                FlushLiteral();
                FormatItem tmp;
                tmp.type   = FormatItem::String;
                tmp.length = -1;
                format.push_back(tmp);
                spos += 1; continue;
            }
            if(c == '!')
            {
                FlushLiteral();
                FormatItem tmp;
                tmp.type   = FormatItem::String;
                tmp.length = 1;
                format.push_back(tmp);
                spos += 1; continue;
            }
            if(c == '\\' && remain >= 2)
            {
                int strlen = 2;
                for(size_t p=spos+1; p<slen; ++p)
                {
                    if(s[p] == '\\') break;
                    if(s[p] != ' ') goto not_valid_backslash_thing;
                    ++strlen;
                }

                { FlushLiteral();
                FormatItem tmp;
                tmp.type   = FormatItem::String;
                tmp.length = strlen;
                format.push_back(tmp);
                spos += strlen; continue;
                }
            not_valid_backslash_thing:;
            }

            if(1) // numeric processing
            {
                size_t saved_spos = spos;

                #define RequireNextC(n) \
                    do { spos += n; if(spos >= slen) goto IllCall; \
                       } while(0)
                #define AcquireNextC(n) \
                    do { spos += n; if(spos >= slen) goto DoneNumFmt; \
                       } while(0)

                // be optimistic, start creating the numeral structure.
                FormatItem options_definitelynumeric;
                FormatItem options_maybenumeric;
                options_definitelynumeric.numdata = 0; // fill all numeric data by zero.
                options_maybenumeric.numdata = 0;

                if(c == '+') // heading +
                {
                    options_maybenumeric.sign_type         = FormatItem::AlwaysSignLeft;
                    options_maybenumeric.num_leading_chars += 1;
                    RequireNextC(1);
                }
                if(c == '$' && remain >= 2 && s[spos+1]=='$')
                {
                    options_definitelynumeric.leading_dollar    = true;
                    options_definitelynumeric.num_leading_chars += 2;
                    AcquireNextC(2);
                }
                else if(c == '*' && remain >= 2 && s[spos+1]=='*')
                {
                    int n = 2;
                    if(remain >= 3 && s[spos+2]=='$')
                    {
                        options_definitelynumeric.leading_dollar    = true;
                        n = 3;
                    }
                    options_definitelynumeric.fill_stars        = true;
                    options_definitelynumeric.num_leading_chars += n;
                    AcquireNextC(n);
                }
                for(;;)
                {
                    if(c == ',' && !options_maybenumeric.decimal_point)
                    {
                        options_maybenumeric.comma_every_third_digit = true;
                        options_maybenumeric.num_leading_chars += 1;
                        AcquireNextC(1);
                    }
                    else if(c == '.' && !options_maybenumeric.decimal_point && !options_definitelynumeric.scientific)
                    {
                        options_maybenumeric.decimal_point = true;
                        AcquireNextC(1);
                    }
                    else if(c == '-' && options_maybenumeric.sign_type != FormatItem::AlwaysSignLeft)
                    {
                        options_maybenumeric.sign_type = FormatItem::NegativeSignRight;
                        AcquireNextC(1);
                        break; // end of the sequence.
                    }
                    else if(c == '+' && options_maybenumeric.sign_type != FormatItem::AlwaysSignLeft)
                    {
                        options_maybenumeric.sign_type = FormatItem::AlwaysSignRight;
                        AcquireNextC(1);
                        break; // end of the sequence.
                    }
                    else if(c == '#' && !options_definitelynumeric.scientific)
                    {
                        if(options_maybenumeric.decimal_point)
                            ++options_definitelynumeric.num_trailing_chars;
                        else
                            ++options_definitelynumeric.num_leading_chars;
                        AcquireNextC(1);
                    }
                    else if(remain >= 4 && c         == '^'
                                        && s[spos+1] == '^'
                                        && s[spos+2] == '^'
                                        && s[spos+3] == '^')
                    {
                        int n = 4;
                        if(remain >= 5 && s[spos+4] == '^') n=5;
                        options_definitelynumeric.scientific = n-2;
                        AcquireNextC(n);
                    }
                    else
                        break;
                }
            DoneNumFmt:;
                //fprintf(stderr, "DoneNumFmt(%s)\n", s+spos);
                if(options_definitelynumeric.numdata != 0) // if something was filled
                {
                    FormatItem tmp;
                    tmp.numdata = options_maybenumeric.numdata | options_definitelynumeric.numdata;

                    if(tmp.num_leading_chars == 0
                    && tmp.num_trailing_chars == 0)
                    {
                        // illegal pattern.
                        spos = saved_spos;
                    }
                    else
                    {
                        FlushLiteral();
                        tmp.type = FormatItem::Numeric;
                        format.push_back(tmp);
                        continue; // accept
                    }
                }
                #undef AcquireNextC
                #undef RequireNextC
            }

            // anything else:
            literal += c;
            spos += 1;

            #undef remain
            #undef c
        }
        FlushLiteral();
      #if 0
        fprintf(stderr, "%u format items\n",(unsigned)format.size());
        for(size_t n=0; n<format.size(); ++n)
        {
            FormatItem& fmt = format[n];
            switch(fmt.type)
            {
                case FormatItem::Literal:
                    fprintf(stderr, "[%d]: Literal \"%s\"\r\n",
                        (int)n, fmt.literal.c_str());
                    break;
                case FormatItem::String:
                    fprintf(stderr, "[%d]: String %d\r\n",
                        (int)n, fmt.length);
                    break;
                case FormatItem::Numeric:
                    fprintf(stderr, "[%d]: Numeric %d.%d sci=%d %s %s %s %s %d\r\n",
                        (int)n,
                        fmt.num_leading_chars,
                        fmt.num_trailing_chars,
                        fmt.scientific,
                        fmt.fill_stars ? "stars" : "no stars",
                        fmt.decimal_point ? "point" : "no point",
                        fmt.comma_every_third_digit ? "comma" : "no comma",
                        fmt.leading_dollar ? "dollar" : "no dollar",
                        (int)fmt.sign_type
                    );
                    break;
               default:
                   fprintf(stderr, "[%d]: eh?\r\n", (int)n);
            }
        }
        fflush(stderr);
      #endif
    }

    static void OutputNonscientific(
        PrintStreamBase& out,
        const BAS_DBL v,
        const UsingFormatter::FormatItem& format)
    {
        std::vector<char> Buffer;

        int bytes_required = snprintf(&Buffer[0], 0,
            "%.*f",
                (int) format.num_trailing_chars,
                std::fabs( v ));
        bytes_required += 8;

        Buffer.resize(bytes_required);

        int begin = 4;

        snprintf(&Buffer[begin], bytes_required - begin,
            "%.*f",
                (int) format.num_trailing_chars,
                std::fabs( v ));

        if(format.leading_dollar) Buffer[--begin] = '$';
        if(format.sign_type == FormatItem::NormalSignLeft && v < 0) Buffer[--begin] = '-';
        if(format.sign_type == FormatItem::AlwaysSignLeft) Buffer[--begin] = (v < 0 ? '-' : '+');

        std::string num_basis = &Buffer[begin];
        if(format.sign_type == FormatItem::NegativeSignRight) num_basis += (v < 0 ? '-' : ' ');
        if(format.sign_type == FormatItem::AlwaysSignRight) num_basis += (v < 0 ? '-' : '+');

        if(format.comma_every_third_digit)
        {
            std::string::size_type decimpoint = num_basis.find('.');
            if(decimpoint == num_basis.npos) decimpoint = num_basis.size();

            while(decimpoint > 3)
            {
                if(std::isdigit(num_basis[decimpoint-1])
                && std::isdigit(num_basis[decimpoint-2])
                && std::isdigit(num_basis[decimpoint-3])
                && std::isdigit(num_basis[decimpoint-4])
                  )
                {
                    num_basis.insert(decimpoint-3, 1, ',');
                    decimpoint -= 3;
                    continue;
                }
                break;
            }
        }

        std::string::size_type decimpoint = num_basis.find('.');
        if(decimpoint == num_basis.npos) decimpoint = num_basis.size();

        if(decimpoint == std::size_t(format.num_leading_chars+1)
        && num_basis[0]=='-'
        && num_basis[1]=='0'
        && num_basis[2]=='.')
        {
            --decimpoint;
            num_basis.erase(1,1);
        }

        if(decimpoint > format.num_leading_chars)
        {
            num_basis.insert(0, 1, '%'); // indicate overflow
        }
        else
        {
            num_basis.insert(0, format.num_leading_chars-decimpoint,
                                format.fill_stars ? '*' : ' ');
        }

        out << num_basis;
    }

private:
    UsingFormatter(const UsingFormatter&);
    void operator=(const UsingFormatter&);
};

)";

static const std::string Stream_Common = R"(
}
#include <map>
#include <memory> // shared_ptr
#include <sstream>
namespace
{
    class InputStreamBase
    {
    public:
        InputStreamBase() {}
        virtual ~InputStreamBase() {}

        // LINE INPUT (console and file)
        virtual void ReadLine(std::string& target) = 0;

        // INPUT$  (console and file)
        virtual std::string Input_N(size_t n) = 0;

        virtual void BeginInput() = 0;
        virtual void InputTo(BAS_INT& target) = 0;
        virtual void InputTo(BAS_LNG& target) = 0;
        virtual void InputTo(BAS_SGN& target) = 0;
        virtual void InputTo(BAS_DBL& target) = 0;
        virtual void InputTo(std::string& target) = 0;
        virtual bool EndInputOkp() = 0; // end input, were all fields input?
        /* In console input:
             At call of EndInputOkp, false is returned
             if any of the InputTos caused error or if
             line was not completely emptied by InputTos
           In file input:
             InputTo may span multiple lines, half-read
             line does not matter. EndInputOkp returns false
             only if an InputTo was unsatisfied.
         */

        static inline const char* GetRedoMsg() { return "\n?Redo from start\n"; }

    private:
        InputStreamBase(const InputStreamBase&);
        void operator=(const InputStreamBase&);
    };

    class InputImplement: public InputStreamBase
    {
    protected:
        bool InputError;
        bool GotEOL;
    public:
        virtual void InputTo(BAS_INT& target)
        {
            long l = InputLong(); target = l;
            if(l < -32768 || l > 32767) GotOverflow();
        }
        virtual void InputTo(BAS_SGN& target)
        {
            target = InputLong();
        }
        virtual void InputTo(BAS_DBL& target)
        {
            target = InputLong();
        }
        virtual void InputTo(BAS_LNG& target)
        {
            long double l = InputLong(); target = l;
            if(l < -2147483648.0
            || l > 2147483647.0) GotOverflow();
        }
        virtual void InputTo(std::string& target)
        {
            // anything until comma or eol or eof
            // ignore spaces at beginning and end
            // if quote encountered while string empty, parse until quote or eol or eof.
            //    After quote, ignore spaces.
            target = "";
            bool quotemode = false;
            bool hadquote  = false;
            for(;;)
            {
                int ch = InputChar();
                if(ch == -1) break;
                if((target.empty() || hadquote) && (ch == ' ' || ch == '\t')) continue;

                /*fprintf(stderr, "gotc(%c)\n", ch);*/

                if(ch == '"' && target.empty()) { quotemode = true; continue; }
                if(ch == '"' && quotemode) { quotemode=false; hadquote=true; continue; }
                if(ch == ',' && !quotemode) break;
                if(hadquote) { InputError = true; break; }
                target += (char)ch;
            }
            std::size_t p=0;
            while(p < target.size() && (target[p]==' ' || target[p]=='\t')) ++p;
            target.erase(target.begin(), target.begin()+p); // LTRIM_S

            /*fprintf(stderr, "target(%s)err(%d)\n",
                target.c_str(), (int)InputError);*/
        }

    protected:
        virtual int InputChar() = 0;
        virtual void GotOverflow() = 0;

        long double InputLong()
        {
            // anything until comma or eol or eof
            // ignore spaces at beginning and end

            std::string target;
            for(;;)
            {
                int ch = InputChar();
                if(ch == -1) break;
                if(target.empty() && (ch == ' ' || ch == '\t')) continue;

                if(ch == ',') break;
                target += (char)ch;
            }

            char* endptr;
            const char* beginptr = target.c_str();
            long double res = strtold(beginptr, &endptr);
            /*fprintf(stderr, "res(%lf)begin(%s)end(%s)\n",
                (double)res,
                beginptr,
                endptr);*/

            if(endptr == beginptr) { InputError = true; return 0.0; }
            while(*endptr)
            {
                if(*endptr != ' ' && *endptr != '\t') { InputError = true; break; }
                ++endptr;
            }
            return res;
        }
        bool FinishInput()
        {
            while(!GotEOL)
            {
                int ch = InputChar();
                if(ch == -1) break;
                if(ch != ' ' || ch != '\t')
                {
                    //fprintf(stderr, "Finish: '%c'?\n", ch);
                    InputError = true; return false;
                }
            }
            return InputError;
        }
    };

    /* File input */
    class QbFIstream: public InputImplement
    {
    public:
        FILE* fp;
        std::vector<char> buffer;
    public:
        QbFIstream(int fd, unsigned reclen=0) : fp(fd < 0 ? nullptr : fdopen(fd, "r"))
        {
            if(fd < 0) return;
            if(!fp) std::perror("fdopen(\"r\"\051"); // \051 codes for ')'
            else if(reclen) { buffer.resize(reclen); std::setvbuf(fp, &buffer[0], _IOFBF, reclen); }
        }

        virtual void ReadLine(std::string& target)
        {
            target = "";
            for(;;)
            {
                int c = InputChar();
                if(c == '\r') continue;
                if(c == '\n' || c == -1) break;
                target += (char) c;
            }
        }

        virtual void BeginInput()
        {
            InputError = false;
            GotEOL     = false;
            // File:
            //   Do nothing. InputTo reads file.
        }

        virtual bool EndInputOkp()
        {
            // File:
            //   Return false if InputTo failed, true otherwise.
            if(InputError) return false;
            return !FinishInput();
        }

        virtual std::string Input_N(size_t n)
        {
            std::string buf(n, '\0');
            size_t res = fread(&buf[0], 1, n, fp);
            if(res == 0)
            {
                if(ferror(fp)) { clearerr(fp); B_ERROR_Errno(); }
            }
            buf.erase(buf.begin() + res, buf.end());
            return buf;
        }

        virtual int InputChar()
        {
         retry:;
            int res = fgetc(fp);
            if(res == '\r') goto retry;
            if(res == EOF || res == '\n') { GotEOL = true; return -1; }
            return res;
        }

        virtual void GotOverflow()
        {
            B_ERROR(6); // overflow
            InputError = true;
        }
    };

    struct PrintCRLFtag { };
    struct PrintSPCtag  { int count; };
    struct PrintTABtag  { int count; };
    struct PrintNextTABtag { };

    class PrintStreamBase
    {
    public:
        PrintStreamBase() {}
        virtual ~PrintStreamBase() {}

        PrintStreamBase& operator<< (BAS_INT v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (BAS_UINT v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (BAS_BYTE v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (BAS_UBYTE v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (BAS_ULONG v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (BAS_LNG v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (BAS_SGN v)
        {
            std::ostringstream s; s << ' ' << v; WriteWithFitCheck( s.str() ); return *this;
        }
        PrintStreamBase& operator<< (BAS_DBL v)
        {
            std::ostringstream s; s << ' ' << v; WriteWithFitCheck( s.str() ); return *this;
        }
        PrintStreamBase& operator<< (const std::string& v)
        {
            WriteWithFitCheck(v); return *this;
        }
        PrintStreamBase& operator<< (long long int v)
        {
            WriteWithFitCheck( std::to_string(v) ); return *this;
        }
        PrintStreamBase& operator<< (const char* v)
        {
            WriteWithFitCheck(v); return *this;
        }
        PrintStreamBase& operator<< (char v)
        {
            WriteWithFitCheck(&v, 1); return *this;
        }

        inline PrintStreamBase& operator<< (PrintCRLFtag)    { DoCRLF();       return *this; }
        inline PrintStreamBase& operator<< (PrintNextTABtag) { DoNextTAB();    return *this; }
        inline PrintStreamBase& operator<< (PrintSPCtag t)   { DoSPC(t.count); return *this; }
        inline PrintStreamBase& operator<< (PrintTABtag t)   { DoTAB(t.count); return *this; }

        virtual void Write(const void* buf, size_t n) = 0;

        // DoCRLF is not the same as either '\r' nor '\n'.
        // Only on terminal, '\r' and '\n' both produce the same as DoCRLF().
        // In files, DoCRLF() produces '\r\n' whereas CHR$ produces individual bytes.
        virtual void DoCRLF() = 0;

        virtual unsigned GetXpos() const = 0;  // 0..n-1
        virtual unsigned GetXsize() const = 0; // n. 0 = infinite (files)

        // SPC() actually produces space characters. It does not only move the cursor.
        void DoSPC(int count)
        {
            if(count < 0) count = 0;

            unsigned xsize = GetXsize();
            if(xsize) count %= xsize;

            std::string spaces(count, ' ');
            Write(&spaces[0], spaces.size());
        }
        void DoTAB(int newpos)
        {
            --newpos;
            if(newpos < 0) newpos = 0;
            int xpos = GetXpos();
            unsigned xsize = GetXsize();
            if(xsize) newpos %= xsize;

            if(newpos < xpos) { xpos=newpos=0; DoCRLF(); }
            if(newpos > xpos) Write(std::string(newpos-xpos, ' '));
        }
        void DoNextTAB()
        {
            const unsigned ColumnWidth = 14;
            static const char SpaceBuf[ColumnWidth+1] = "              "; // 14 spaces
            int xpos = GetXpos();
            int columnpos   = xpos % ColumnWidth;
            int need_spaces = ColumnWidth - columnpos;
            if(CheckFit(need_spaces))
                Write(SpaceBuf, need_spaces);
        }
    private:
        bool CheckFit(int n_columns) // no const, it may print CRLF
        {
            int xsize = GetXsize(); if(!xsize) return true;
            int xpos = GetXpos();
            if(xpos+n_columns < xsize) return true;
            if(xpos > 0) DoCRLF();
            return false;
        }
        void WriteWithFitCheck(const void* buf, std::size_t n)
        {
            CheckFit(n);
            Write(buf, n);
        }
        void WriteWithFitCheck(const std::string& s)
        {
            WriteWithFitCheck(s.data(), s.size());
        }
        void Write(const std::string& s)
        {
            Write(s.data(), s.size());
        }

    private:
        PrintStreamBase(const PrintStreamBase&);
        void operator=(const PrintStreamBase&);
    };

    /* File writing */
    class QbFPstream: public PrintStreamBase
    {
    public:
        FILE* fp;
        int xpos;
        std::vector<char> buffer;
    public:
        QbFPstream(int fd, unsigned reclen=0) : fp(fd < 0 ? nullptr : fdopen(fd, "w")), xpos(0)
        {
            if(fd < 0) return;
            if(!fp) std::perror("fdopen(\"w\"\051"); // \051 codes for ')'
            else if(reclen) { buffer.resize(reclen); std::setvbuf(fp, &buffer[0], _IOFBF, reclen); }
        }

        virtual unsigned GetXpos() const
        {
            return xpos;
        }
        virtual unsigned GetXsize() const
        {
            return 65535;
        }
        virtual void DoCRLF()
        {
            fputc('\n', fp);
            fflush(fp);
        }
        virtual void Write(const void* buf, size_t n)
        {
            size_t res = fwrite(buf, 1, n, fp);
            if(!res) B_ERROR_Errno();
            fflush(fp);
        }
    };

    struct FileDescriptor
    {
        int fd; // system filedescriptor
        QbFPstream fpout;
        QbFIstream fpin;

        std::size_t record_width;

        struct Field
        {
            std::size_t offset;
            std::size_t width;
            std::string* target;
        };

        std::vector<Field> fields;

        bool readable;
        bool writable;

    public:
        FileDescriptor(int f, unsigned reclen, bool r,bool w)
            : fd(f),
              fpout(w ? f : -1, reclen),
              fpin(r ? f : -1, reclen),
              record_width(reclen),
              fields(), readable(r), writable(w)
        {
        }
    };
    std::map<int/*qbasic filenum*/, std::shared_ptr<FileDescriptor>> fdmap;

    extern PrintStreamBase& GET_PRINTER();

    InputStreamBase& GET_STREAM_INPUT(int filenumber)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        if(!i->second->readable) B_ERROR(54); // bad file mode
        return i->second->fpin;
    }

    PrintStreamBase& GET_STREAM_OUTPUT(int filenumber)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        if(!i->second->writable) B_ERROR(54); // bad file mode

        return i->second->fpout;
    }
)";

static const std::string Console_Common = R"(
}
#include <unistd.h> // isatty
namespace
{
    /* Console I/O */
    class QbConsoleStream: public PrintStreamBase, public InputImplement
    {
        std::string InputLineBuffer;
        unsigned InputLineBufferPos;
    public:
        QbConsoleStream() : PrintStreamBase(), InputImplement() { }

        virtual unsigned GetXpos() const
        {
            return B_FuncPOS();
        }
        virtual unsigned GetXsize() const
        {
            return B_FuncWIDTH();
        }
        virtual void DoCRLF()
        {
            GFX_OutS("\r", 1);
        }
        virtual void Write(const void* buf, size_t n)
        {
            GFX_OutS( (const char*) buf, n);
        }

        virtual void ReadLine(std::string& target)
        {
            target = "";
            std::size_t strpos = 0;
            int curspos = B_FuncPOS();
            bool insertmode = true;
            for(;;)
            {
                std::string inkey = B_FuncINKEY_S(true);
                if(inkey == "\n" || inkey == "\r") break;
                if(inkey[0] == '\0')
                    switch(inkey[1])
                    {
                        case 'K': // left
                            if(strpos > 0) { --strpos; --curspos; goto relocate; }
                            break;
                        case 'M': // right
                            if(strpos < target.size())
                                           { ++strpos; ++curspos; goto relocate; }
                            break;
                        case 'O': // end
                            curspos += target.size() - strpos;
                            strpos = target.size(); goto relocate;
                        case 'Q': // home
                            curspos -= strpos; strpos = 0; goto relocate;
                        default: continue;
                    }
                else if(inkey.size() == 1)
                    switch(inkey[0])
                    {
                        case 8: // backspace
                            if(strpos > 0)
                                { target.erase(--strpos, 1); --curspos;
                                  B_StmtLOCATE(-1, curspos);
                                  GFX_OutS(&target[strpos], target.size()-strpos);
                                  GFX_OutS(" ", 1);
                                  goto relocate;
                                }
                            continue;
                    }

                target.insert(strpos, inkey);

                GFX_OutS(&target[strpos], target.size()-strpos);

                if(insertmode)
                {
                    strpos += inkey.size();
                    curspos += inkey.size();
                }
                else
                    target.erase(strpos + inkey.size(), inkey.size());
            relocate:
                B_StmtLOCATE(-1, curspos);
            }

            if(strpos != target.size())
            {
                curspos += target.size() - strpos;
                strpos = target.size();
                B_StmtLOCATE(-1, curspos);
            }
            // note: does not go to next line.
        }

        virtual void BeginInput()
        {
            InputError = false;
            GotEOL     = false;
            // Console:
            //   Read line into buffer. InputTo queries buffer.
            ReadLine(InputLineBuffer); // note: keeps cursor on same line
            InputLineBufferPos=0;
        }

        virtual bool EndInputOkp()
        {
            // Console:
            //   Return false if InputTo failed.
            //   Return false if buffer is not empty.
            //   Return true otherwise.

            if(InputError) return false;
            return !FinishInput();
        }

        virtual std::string Input_N(size_t n)
        {
            // No echoing
            // FIXME: Don't take control chars.
            std::string res;
            for(; n > 0; --n) res += B_FuncINKEY_S(true);
            return res;
        }

        virtual int InputChar()
        {
            if(InputLineBufferPos < InputLineBuffer.size())
            {
                return InputLineBuffer[InputLineBufferPos++];
            }
            GotEOL = true;
            return -1;
        }

        virtual void GotOverflow()
        {
            InputError = true;
            *this << "Overflow\n";
        }
    } cons;

    InputStreamBase& GetStdin()  { static QbFIstream in(0); return in; }
    InputStreamBase& GET_STDIN() { static InputStreamBase& result = isatty(0) ? cons : GetStdin(); return result; }
    PrintStreamBase& GetStdout() { static QbFPstream out(1); return out; }
    PrintStreamBase& GET_STDOUT() { static PrintStreamBase& result = isatty(1) ? cons : GetStdout(); return result; }
)";


static const std::string DYNARRAY_Common = R"(
    template<typename T, unsigned numdims>
    class BAS_DYNARRAY
    {
    public:
        template<typename... U>
        BAS_DYNARRAY(U... dimensions)
        {
            Redim_Sizes(0,0, dimensions...);
        }

        void Redim_Sizes(unsigned,std::size_t total_size)
        {
            container.resize(total_size);
        }

        template<typename... U>
        void Redim_Sizes(unsigned which_index,std::size_t total_size, int min, std::size_t width, U... dimensions)
        {
            length[which_index] = width;
            begin[which_index]  = min;
            if(which_index == 1)
                factor[which_index-1] = length[which_index-1];
            else if(which_index > 0)
                factor[which_index-1] = factor[which_index-2] * length[which_index-1];
            Redim_Sizes(which_index+1, total_size*width, dimensions...);
        }

        void Redim(unsigned,std::size_t total_size)
        {
            container.resize(total_size);
        }

        template<typename... U>
        void Redim(unsigned which_index,std::size_t total_size, int min, int max, U... dimensions)
        {
            std::size_t width = max-min+1;

            length[which_index] = width;
            begin[which_index]  = min;
            if(which_index == 1)
                factor[which_index-1] = length[which_index-1];
            else if(which_index > 0)
                factor[which_index-1] = factor[which_index-2] * length[which_index-1];
            Redim(which_index+1, total_size*width, dimensions...);
        }

        void Erase()
        {
            for(unsigned n=0; n<numdims; ++n) { length[n]=0; begin[n]=0; }
            for(unsigned n=0; n<numdims-1; ++n) { factor[n]=0; }
            container.clear();
        }

        int Lbound(int which_index) const __attribute__((regparm(2)))
        {
            std::size_t i = which_index - 1; // unsigned result
            if(unlikely(i >= numdims)) B_ERROR(9); // subscript out of range
            return begin[i];
        }
        int Ubound(int which_index) const __attribute__((regparm(2)))
        {
            std::size_t i = which_index - 1; // unsigned result
            if(unlikely(i >= numdims)) B_ERROR(9); // subscript out of range
            return begin[i] + length[i] - 1;
        }
    private:
        T& Index(std::size_t start,unsigned) __attribute__((regparm(6)))
        {
            //if(unlikely(start >= container.size())) B_ERROR(9);
            return container[start];
        }
        const T& Index(std::size_t start,unsigned) const __attribute__((regparm(6)))
        {
            //if(unlikely(start >= container.size())) B_ERROR(9);
            return container[start];
        }
    public:
        template<typename... U>
        inline T& Index(std::size_t start, unsigned which_index, int index, U... rest)
        {
            std::size_t offs = index - begin[which_index]; // unsigned result
            if(unlikely(offs >= length[which_index])) B_ERROR(9);
            if(which_index != 0) offs *= factor[which_index-1];
            return Index(start + offs, which_index+1, rest...);
        }

        template<typename... U>
        const inline T& Index(std::size_t start, unsigned which_index, int index, U... rest) const
        {
            std::size_t offs = index - begin[which_index]; // unsigned result
            if(unlikely(offs >= length[which_index])) B_ERROR(9);
            if(which_index != 0) offs *= factor[which_index-1];
            return Index(start + offs, which_index+1, rest...);
        }

    private:
        void RedimUpdate();
    public:
        // Should be private, but I can't figure out the syntax
        // to "friend" these to BAS_MAGICARRAY<T>::DynWrapper<numdims>
        std::size_t length[numdims];
        std::size_t factor[numdims > 1 ? numdims-1 : 1];
        int         begin[numdims];
        std::vector<T> container;
    };
)";

static const std::string MAGICARRAY_Common = R"(
}
#include <memory> // shared_ptr
namespace
{
    /* Magical arrays are arrays that have the following properties:
     * They are arrays and they have a type.
     *
     *
     * Any other aspects are runtime variants:
     *
     *   Number of dimensions?
     *   Initialized by a static array?
     *   Initialized by a dynamic array?
     *   Initialized by another magic array? - Not supported by QB
     *   Minimums and maximums on each dimension
     *
     *   REDIM is not allowed on a magic array, so at least there's that.
     *   But ERASE is.
     */

    template<unsigned N>
    struct BAS_MAGICARRAY_SIZETAG { };

    template<typename T>
    class BAS_MAGICARRAY
    {
        class MagicBase
        {
        public:
            virtual ~MagicBase() { }
            virtual void Erase() = 0;
            virtual int Lbound(int which_index) const = 0;
            virtual int Ubound(int which_index) const = 0;

            virtual T& Index(std::initializer_list<long>&& indexes) = 0;
            virtual const T& Index(std::initializer_list<long>&& indexes) const = 0;
        };

        template<unsigned numdims>
        class DynWrapper: public MagicBase
        {
        public:
            DynWrapper(BAS_DYNARRAY<T,numdims>& d) : data(d) { }

            inline ~DynWrapper() {}

            virtual void Erase()
            {
                data.Erase();
            }

            virtual int Lbound(int which_index) const
            {
                return data.Lbound(which_index);
            }
            virtual int Ubound(int which_index) const
            {
                return data.Ubound(which_index);
            }

            virtual T& Index(std::initializer_list<long>&& indexes)
            {
                std::size_t start       = 0;
                unsigned    which_index = 0;
                if(unlikely(indexes.end() - indexes.begin() != numdims)) B_ERROR(9); // subscript out of range
                for(int index: indexes)
                {
                    std::size_t offs = index - data.begin[which_index]; // unsigned result
                    if(which_index != 0) offs *= data.factor[which_index-1];
                    start += offs;
                    ++which_index;
                }
                if(unlikely(start >= data.container.size())) B_ERROR(9); // subscript out of range
                return data.container[start];
            }

            virtual const T& Index(std::initializer_list<long>&& indexes) const
            {
                std::size_t start       = 0;
                unsigned    which_index = 0;
                if(unlikely(indexes.end() - indexes.begin() != numdims)) B_ERROR(9); // subscript out of range
                for(int index: indexes)
                {
                    std::size_t offs = index - data.begin[which_index]; // unsigned result
                    if(which_index != 0) offs *= data.factor[which_index-1];
                    start += offs;
                    ++which_index;
                }
                if(unlikely(start >= data.container.size())) B_ERROR(9); // subscript out of range
                return data.container[start];
            }
        private:
            BAS_DYNARRAY<T,numdims>& data;
        };

        template<unsigned numdims>
        class StaticWrapper: public MagicBase
        {
        public:
            template<typename... U>
            StaticWrapper(T* d, U... dimensions) : data(d)
            {
                totalsize = InitDimensions(0, dimensions...);
            }

            inline ~StaticWrapper() {}

            virtual void Erase()
            {
                for(std::size_t a=0; a<totalsize; ++a) data[a] = T();
            }

            virtual int Lbound(int which_index) const
            {
                std::size_t i = which_index - 1; // unsigned result
                if(unlikely(i >= numdims)) B_ERROR(9); // subscript out of range
                return begin[i];
            }
            virtual int Ubound(int which_index) const
            {
                std::size_t i = which_index - 1; // unsigned result
                if(unlikely(i >= numdims)) B_ERROR(9); // subscript out of range
                //fprintf(stderr, "STATIC Begin[%u]=%X, length[%u]=%X, Ubound=%X\n", int(i),int(begin[i]),int(i),int(length[i]),int(begin[i]+length[i]-1));
                return begin[i] + length[i] - 1;
            }

            virtual T& Index(std::initializer_list<long>&& indexes)
            {
                std::size_t start       = 0;
                unsigned    which_index = 0;
                for(int index: indexes)
                {
                    std::size_t offs = index - begin[which_index]; // unsigned result
                    if(which_index != 0) offs *= factor[which_index-1];
                    start += offs;
                    ++which_index;
                }
                if(unlikely(which_index != numdims || start >= totalsize)) B_ERROR(9); // subscript out of range
                return data[start];
            }

            virtual const T& Index(std::initializer_list<long>&& indexes) const
            {
                std::size_t start       = 0;
                unsigned    which_index = 0;
                for(int index: indexes)
                {
                    std::size_t offs = index - begin[which_index]; // unsigned result
                    if(which_index != 0) offs *= factor[which_index-1];
                    start += offs;
                    ++which_index;
                }
                if(unlikely(which_index != numdims || start >= totalsize)) B_ERROR(9); // subscript out of range
                return data[start];
            }

        private:
            static inline std::size_t InitDimensions(unsigned)
            {
                return 1;
            }

            template<typename... U>
            inline std::size_t InitDimensions(unsigned which_index, int min, std::size_t width, U... dimensions)
            {
                length[which_index] = width;
                begin[which_index]  = min;
                if(which_index == 1)
                    factor[which_index-1] = length[which_index-1];
                else if(which_index > 0)
                    factor[which_index-1] = factor[which_index-2] * length[which_index-1];
                return width * InitDimensions(which_index+1, dimensions...);
            }
        private:
            T* data;
            std::size_t length[numdims];
            std::size_t factor[numdims > 1 ? numdims-1 : 1];
            int         begin[numdims];
            std::size_t totalsize;
        };

    public:
        // Initializing with a static array:
        template<unsigned N, typename... U>
        BAS_MAGICARRAY(T* data, BAS_MAGICARRAY_SIZETAG<N>, U... dimensions)
            : impl( std::make_shared<StaticWrapper<N>>(data, dimensions...) )
        {
        }

        // Initializing with a dynamic array:
        template<unsigned numdims>
        BAS_MAGICARRAY(BAS_DYNARRAY<T,numdims>& data)
            : impl( std::make_shared<DynWrapper<numdims>> (data) )
        {
        }

        // Initializing with another magic array:
        BAS_MAGICARRAY(const BAS_MAGICARRAY<T>& b)
            : impl(b.impl)
        {
        }

        T& Index(std::initializer_list<long>&& indexes)
        {
            return impl->Index( std::move(indexes) );
        }

        const T& Index(std::initializer_list<long>&& indexes) const
        {
            return impl->Index( std::move(indexes) );
        }

        void Erase()                     { impl->Erase(); }
        int Lbound(int dimindex=1) const { return impl->Lbound(dimindex); }
        int Ubound(int dimindex=1) const { return impl->Ubound(dimindex); }

        // REDIM is not permitted for parameter arrays
        // void Redim(unsigned numdims, ...);

        BAS_MAGICARRAY<T>& operator=(const BAS_MAGICARRAY<T>&) = delete;

    private:
        std::shared_ptr<MagicBase> impl;
    };
)";
