tokcase(LSET, "V=S",  R"(
    void B_StmtLSET(std::string& result, const std::string& input)
    {
        std::size_t copy_length = std::min( result.size(), input.size() );
        std::size_t pad_length  = std::max( result.size(), input.size() );

        result.replace(0, copy_length, input, 0, copy_length);
        result.replace(copy_length, pad_length-copy_length, pad_length-copy_length, ' ');
    }
)");
tokcase(RSET, "V=S",   R"(
    void B_StmtRSET(std::string& result, const std::string& input)
    {
        std::size_t copy_length = std::min( result.size(), input.size() );
        std::size_t pad_length  = std::max( result.size(), input.size() );

        result.replace(pad_length-copy_length, copy_length, input, 0, copy_length);
        result.replace(0, pad_length-copy_length, pad_length-copy_length, ' ');
    }
)");
tokcase(MID_S,    "(V,N?,N)=S",   R"(
    void B_StmtMID_S(std::string& target, int offset, int number, const std::string& input)
    {
        if(offset <= 0 || number < 0 || std::size_t(offset+number) > target.size()) B_ERROR(5); // illegal function call
        target.replace(offset-1, number, input);
    }
)");
tokcase(ASC,      "(V,N)=N",      R"(
    void B_StmtASC(std::string& s, int offset, int n) // MID$(s, offset, 1) = CHR$(value) (qb64 extension)
    {
        std::size_t p = offset-1; // produces unsigned
        if(p >= s.size()) B_ERROR(5); // illegal function call
        if(n < 0 || n > 255) B_ERROR(5); // illegal function call
        s[p] = (char)n;
    }
)");
tokcase(SWAP,     "V,V",         R"(
    template<typename T, typename U>
    static inline void B_StmtSWAP(T& a, U& b)
    {
        std::swap(a, b);
    }
)");
     /* system commands */
tokcase(STOP,     "",  "");
tokcase(END,      "",    R"(
    void B_StmtEND(int code = 0)    __attribute__((noreturn));
    void B_StmtEND(int code)
    {
        B_LibDone();
        std::exit(code);
    }
)");
tokcase(SYSTEM,   "?N",   R"(
    void B_StmtSYSTEM(int code = 0) __attribute__((noreturn));
    void B_StmtSYSTEM(int code)
    {
        B_LibDone();
        std::exit(code);
    }
)");
tokcase(SHELL,    "s", "");
tokcase(SLEEP,    "n", "");
tokcase(POKE,     "N,N", "");
tokcase(OUT,      "N,N", "");
tokcase(WAIT,     "N,N,n", "");
tokcase(DEFSEG,   "?=N",   R"(
    void B_StmtDEFSEG(int seg=-1)
    {
        B_DefSeg = seg;
    }
)");
tokcase(TIME_S,   "=S",  ( std::initializer_list<std::string>{ TIMES_DATES_Common, Func_DATES_Common, R"(
    void B_StmtTIME_S(const std::string& t)
    {
        // accepted formats: hh, hh:mm, hh:mm:ss
        TrySetTime(B_FuncDATE_S() + " " + t); // date + time
    }
)"} ) );
tokcase(DATE_S,   "=S",  ( std::initializer_list<std::string>{ TIMES_DATES_Common, Func_TIMES_Common, R"(
    void B_StmtDATE_S(const std::string& d)
    {
        // accepted formats:  mm-dd-yy, mm-dd-yyyy, mm/dd/yy, mm/dd/yyyy
        TrySetTime(d + " " + B_FuncTIME_S()); // date + time
    }
)"} ) );

tokcase(CLEAR,    "n?,n?,n",   R"(
    void B_CLEAR(int ds_size,int,int ss_size, bool in_sub)
    {
        if(in_sub)
        {
            B_ERROR(5);
            // QuickBASIC throws "illegal function call"
            // if CLEAR called from inside a SUB. Not sure why.
        }

        ds_size=ds_size; // ignored
        ss_size=ss_size; // ignored

        GosubStack.clear(); // Defined to do so in QB manual

        B_StmtCLOSE(); // Close all open files

        ON_ERROR_GOTO_NOWHERE();

        /* This clears possible active error. */
        if(B_ERR_VALUE)
        {
            B_RESUME_TO(cont);
        cont:;
        }
    }
)");
tokcase(BEEP,     "",          R"(
    void B_StmtBEEP() { B_StmtSOUND(400, 6); }
)");
tokcase(NOISE,    "n,n,n", "");
tokcase(SOUND,    "n,n", "");

tokcase(KILL,     "S", "");
tokcase(FILES,    "s", "");
tokcase(SEEK,     "F,N", "");
tokcase(CHDIR,    "S", "");
tokcase(MKDIR,    "S", "");
tokcase(RMDIR,    "S", "");
tokcase(ERROR,    "N", "");

     /* graphics commands */
tokcase(LOCATE,   "n?,n?,n?,n?,n", "");  // row,col, cursor_{vis,start,stop}
tokcase(SCREEN,   "n?,n?,n?,n", "");     // mode,colorswitch,apage,vpage
tokcase(COLOR,    "n?,n?,n", "");        // fg,bg,border
tokcase(PAINT,    "P,S?,n?,s"
                  "|"
                  "P?,n?,n", ""); // coord, pattern/paint, bordercolor, background
tokcase(CIRCLE,   "P,N?,n?,n?,n?,n", ""); // coord, radius, color, start, end, aspect
tokcase(PSET,     "P?,n", ""); //coord,color
tokcase(PRESET,   "P?,n", "");
tokcase(CLS,      "n", "");   //clear_level (0/1/2)
tokcase(DRAW,     "S", "");
tokcase(GET,      "P-P,A"
                  "|"
                  "F?,n?,V",  ( std::initializer_list<std::string>{ Stream_Common, R"( // File I/O
}
#include <assert.h>
namespace
{
    void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber=-1)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        auto& r = *i->second;
        if(!r.readable) B_ERROR(54);      // bad file mode
        if(r.fields.empty()) B_ERROR(40); // variable required

        // TODO: If file is open in INPUT mode, "bad file mode"
        if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number

        if(recnumber > 0) std::fseek(r.fpin.fp, (recnumber-1) * r.record_width, SEEK_SET);

        std::vector<char> buffer(r.record_width);

        if(std::fread(&buffer[0], 1, buffer.size(), r.fpin.fp) == 0)
        {
            if(std::ferror(r.fpin.fp))
                B_ERROR_Errno();
            else
                B_ERROR(62); // input past end of file
        }

        for(const auto& field: r.fields)
        {
            /*fprintf(stderr, "Field offset %u, width %u, buffer is %u, target (%p) is %u(%s)\n",
                (unsigned)field.offset, (unsigned)field.width,
                (unsigned)buffer.size(), field.target, (unsigned)field.target->size(), field.target->c_str());*/

            assert(field.offset + field.width <= buffer.size());
            assert(field.width <= field.target->size());
            field.target->assign( &buffer[field.offset], field.width );
        }
    }

    void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, void* target, std::size_t bytes)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        auto& r = *i->second;
        if(!r.readable) B_ERROR(54);      // bad file mode

        if(!r.fields.empty()) B_ERROR(56); // field statement active

        // TODO: If file is open in INPUT or RANDOM mode, "bad file mode"

        if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number
        if(recnumber > 0) std::fseek(r.fpin.fp, (recnumber-1) * r.record_width, SEEK_SET);

        // TODO: If file is open in RANDOM mode and bytes doesn't match record_width, error

        if(std::fread(target, 1, bytes, r.fpin.fp) == 0)
        {
            if(std::ferror(r.fpin.fp))
                B_ERROR_Errno();
            else
                B_ERROR(62); // input past end of file
        }
    }

    static inline void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, std::string& target)
        { B_StmtGET(filenumber,recnumber, target.empty() ? nullptr : (void*)&target[0], target.size()); }

    template<typename T>
    static inline void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, T& target)
        { B_StmtGET(filenumber,recnumber,(void*)&target, sizeof(target)); }

)" } ) );


tokcase(PUT,      "P,A?,zp"
                  "|"
                  "F?%n?,V",  ( std::initializer_list<std::string>{ Stream_Common, R"( // File I/O
}
#include <assert.h>
namespace
{
    void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        auto& r = *i->second;
        if(!r.writable) B_ERROR(54);      // bad file mode
        if(r.fields.empty()) B_ERROR(40); // variable required

        // TODO: If file is open in OUTPUT mode, "bad file mode"
        if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number
        if(recnumber > 0) std::fseek(r.fpout.fp, (recnumber-1) * r.record_width, SEEK_SET);

        std::vector<char> buffer(r.record_width);

        for(const auto& field: r.fields)
        {
            assert(field.offset + field.width <= buffer.size());
            assert(field.width <= field.target->size());
            std::copy( field.target->begin(),
                       field.target->begin() + field.width,
                       buffer.begin() + field.offset );
        }
        if(std::fwrite(&buffer[0], 1, buffer.size(), r.fpout.fp) == 0) B_ERROR_Errno();
    }

    void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const void* source, std::size_t bytes)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        auto& r = *i->second;
        if(!r.writable) B_ERROR(54);      // bad file mode

        if(!r.fields.empty()) B_ERROR(56); // field statement active

        // TODO: If file is open in INPUT or RANDOM mode, "bad file mode"

        if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number
        if(recnumber > 0) std::fseek(r.fpout.fp, (recnumber-1) * r.record_width, SEEK_SET);

        // TODO: If file is open in RANDOM mode and bytes doesn't match record_width, error

        if(std::fwrite(source, 1, bytes, r.fpout.fp) == 0) B_ERROR_Errno();
    }

    static inline void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const std::string& input)
        { B_StmtPUT(filenumber,recnumber, input.empty() ? nullptr : (const void*)&input[0], input.size()); }

    template<typename T>
    static inline void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const T& source)
        { B_StmtPUT(filenumber,recnumber,(const void*)&source, sizeof(source)); }
)" } ) );

tokcase(VIEW_PRINT,    "N ZT N"
                       "|", "");
tokcase(VIEW,           "zs C-C?,n?,n", ""); // [SCREEN] coord-coord,color,border
tokcase(PALETTE_USING,  "A", "");
tokcase(PALETTE,        "n?,n", "");
tokcase(LINE,           "p-P?,n?,zl?,n", "");  // coord-coord, color, B/BF, style (16-bit intmask)
tokcase(WIDTH_LPRINT, "N", "");
tokcase(WIDTH_F,  "F,N", "");
tokcase(WIDTH,    "S,N"
                  "|"
                  "n?,n", "");
tokcase(WINDOW,   "zs C-C"
                  "|", "");

tokcase(qb64DISPLAY,   "", "");
tokcase(qb64DONTBLEND, "n", "");

     /* event commands */
tokcase(PLAY,     "ZK"
                  "|"
                  "%S", "");
tokcase(TIMER,    "ZK", "");
tokcase(KEY,      "ZH"
                  "|"
                  "N,S", "");

tokcase(RANDOMIZE, "N"
                   "|"     // User will be prompted for random seed
                   "",     ( std::initializer_list<std::string>{ R"(
    void B_RANDOMIZE(int i)
    {
        B_RandomSeed = (B_RandomSeed & ~0xFFFF00u) | ((i & 0xFFFF) << 8);
    }
    void B_StmtRANDOMIZE(BAS_DBL n)
    {
        const std::uint_least16_t* words = (const std::uint_least16_t*) (&n);

        unsigned mask = words[2] ^ words[3];
        B_RANDOMIZE(mask);
    }
    void B_StmtRANDOMIZE()
    {
        /* INPUT "Random Number Seed (-32768 to 32767)?"; seed
           Seed = (Seed & 0xFF0000FFu) | seed
        */
        InputStreamBase& i = GET_STDIN();
        PrintStreamBase& o = GET_STDOUT();
        BAS_INT result = 0;
        for(;;)
        {
            o << "Random Number Seed (-32768 to 32767)? ";
            i.BeginInput();
            i.InputTo(result);
            if(i.EndInputOkp()) break;
            o << i.GetRedoMsg();
        }
        o.DoCRLF();
        B_RANDOMIZE(result);
    }
)" } ) );

    /* file commands */

tokcase(FIELD, "F,%N ZA V", ( std::initializer_list<std::string>{ Stream_Common, R"(
    typedef std::pair<struct FileDescriptor&,std::size_t> B_FIELD_Temp;

    B_FIELD_Temp B_FIELD_Init(BAS_INT filenumber)
    {
        auto i = fdmap.find(filenumber);
        if(i == fdmap.end()) B_ERROR(52); // bad file number
        auto& r = *i->second;
        //r.fields.clear();

        return { r, 0 };
    }

    B_FIELD_Temp B_FIELD(B_FIELD_Temp&& info, unsigned width, std::string& stringvar)
    {
        auto& r = info.first;
        std::size_t record_offset = info.second;
        if(record_offset + width > r.record_width) B_ERROR(50); // field overflow

        stringvar.assign(std::size_t(width), ' ');
        r.fields.emplace_back( FileDescriptor::Field{record_offset, width, &stringvar} );
        /*fprintf(stderr, "Field: assigned (%p) %u(%s)\n",
            &stringvar, (unsigned)stringvar.size(), stringvar.c_str());*/

        return { r, record_offset + width };
    }

    static inline void B_FIELD_Cont(const B_FIELD_Temp&) {}
    template<typename...T>
    static inline void B_FIELD_Cont(B_FIELD_Temp&& info,
                                    unsigned width, std::string& stringvar, T&&... rest)
    {
        B_FIELD_Cont(B_FIELD(std::move(info), width, stringvar), std::forward<T>(rest)...);
    }

    template<typename...T>
    static inline void B_StmtFIELD(BAS_INT filenumber, T&&... rest)
    {
        B_FIELD_Cont(B_FIELD_Init(filenumber), std::forward<T>(rest)...);
    }
)" } ) );

tokcase(CLOSE, "?%F",     ( std::initializer_list<std::string>{ Stream_Common, R"(
    void B_StmtCLOSE()
    {
        for(const auto& f: fdmap) close(f.second->fd);
        fdmap.clear();
    }
    void B_StmtCLOSE(BAS_INT filenumber)
    {
        auto i = fdmap.find(filenumber); if(i == fdmap.end()) B_ERROR(52); // bad file number
        close(i->second->fd);
        fdmap.erase(i);
    }
)" } ) );
    /* Sound commands */
tokcase(qb64SNDBAL, "N?,N?,N?,N", ""); // FIXME: omitted elements should be assumed 0, not -1
tokcase(qb64SNDRAW, "N?,N?,n", "");
tokcase(qb64SNDRAWOPEN, "N", "");
tokcase(qb64SNDRAWDONE, "", "");
tokcase(qb64SNDCLOSE, "N", "");
tokcase(qb64SNDSTOP,  "N", "");
tokcase(qb64SNDPAUSE, "N", "");
tokcase(qb64SNDPLAY,  "N", "");
tokcase(qb64SNDPLAYCOPY, "N?,n", "");

tokcase(READ, "%V",  ( std::initializer_list<std::string>{ READ_VAL_Common, R"(
    template<typename T>
    struct B_DoRead
    {
        B_DoRead(T& target, const char* input)
        {
            B_ReadGeneric<T, true>(target, input);
        }
    };

    template<>
    struct B_DoRead<std::string>
    {
        B_DoRead(std::string& target, const char* input)
        {
            target = input;
        }
    };

    static inline void B_StmtREAD() {}

    template<typename T, typename... Rest>
    void B_StmtREAD(T& target, Rest&&... rest)
    {
        if(unlikely( B_DataPtr >= DATA_COUNT)) { B_ERROR(4); } //OUT OF DATA
        B_DoRead<T>(target, DATA_TABLE[ B_DataPtr++ ]);

        B_StmtREAD( std::forward<Rest>(rest)... );
    }

)"} ) );
