#define PARSER_CLIKE_MAX_CASES 23

#include "expr.hh"
//#include "dimdef.hh"
//#include "casevalue.hh"
#include "binding.hh"
#include "stmt.hh"

#include "lexer.hh"
#include "tokens.hh"

//#define LEXER_DEBUG
//#define PARSER_DEBUG

#include "parser.hh"
#include "parser_lexiterator.hh"

typedef Lex<QBtoken>::iterator iterator_t;
typedef scanner<iterator_t> scanner_t;

#define tk(x) ch_p((QBtoken)x)
#define tq(x) ch_p((QBtoken)x).expected()

#define AP const iterator_t&,const iterator_t&
namespace
{
    exprp option_base = new expr_number(0);
    std::vector<std::string> global_data;

    struct Context
    {
    public:
        exprp exp_lhs;
        exprp exp_result;
        
        stmtp stm_lhs;
        stmtp stm_result;
        
        unsigned count;
        QBtoken token;
        exprp param1, param2, param3, param4, param5;
        stmtp stmt1, stmt2, stmt3;

        autoptr<expr_var> var_lhs, var_result;
    public:
        Context() : count(0),token((QBtoken)0)
        {
        }
    };
    Context ctx;
    std::list<Context> saved_contexts;
    
    void SetExpr(exprp new_res)
    {
        ctx.exp_lhs    = ctx.exp_result;
        ctx.exp_result = new_res;
    }
    
    /////////////////////////////////

    template<typename E>
    struct DoBinaryExpr
    {
        void act(int op) const
        {
            SetExpr(new E( (QBtoken)op, ctx.exp_lhs, ctx.exp_result));
        }
    };
    
    template<typename E, int op>
    struct DoUnaryToggleExpr
    {
        void act(int count) const
        {
            if(count % 2)
                SetExpr(new E( (QBtoken)op, ctx.exp_result));
        }
    };
    
    template<typename E>
    struct DoUnaryExpr
    {
        void act(int op) const
        {
            SetExpr(new E( (QBtoken)op, ctx.exp_result));
        }
    };
    
    template<typename E>
    struct DoZnaryExpr
    {
        void act(int op) const
        {
            SetExpr(new E( (QBtoken)op ));
        }
    };
    
    template<typename E>
    struct Do_2_ParamFunc { void act(int op) const
        { SetExpr(new E( (QBtoken)op, ctx.param1,ctx.param2)); } };
    template<typename E>
    struct Do_3_ParamFunc { void act(int op) const
        { SetExpr(new E( (QBtoken)op, ctx.param1,ctx.param2,ctx.param3)); } };

    /////////////////////////////////

    void DoNumber(AP)
    {
        SetExpr(new expr_number(yylval.number));
    }
    void DoString(AP)
    {
        SetExpr(new expr_string(yylval.str));
    }
    
    /////////////////////////////////

    template<typename T, typename ActionT>
    class int_actor: public ActionT
    {
        T val;
    public:
        explicit int_actor(T v) : val(v) { }
    
        template<typename It>
        void operator() (const It&,const It&) const
        {
            this->act(val);
        }
    };
    template<typename ActionT,typename T>
    int_actor<T,ActionT>
    int_act_a(T p) { return int_actor<T,ActionT>(p); }
    
    /////////////////////////////////

    template<typename T, typename ActionT>
    class int_actor_ref: public ActionT
    {
        const T& val;
    public:
        explicit int_actor_ref(const T& v) : val(v) { }
    
        template<typename It>
        void operator() (const It&,const It&) const
        {
            this->act(val);
        }
    };
    template<typename ActionT,typename T>
    int_actor_ref<T,ActionT>
    int_act_ref_a(T& p) { return int_actor_ref<T,ActionT>(p); }
    
    /////////////////////////////////

    template<typename T>
    class assign_actor_1
    {
        T& val;
    public:
        explicit assign_actor_1(T& v) : val(v) { }
    
        template<typename It>
        void operator() (const It& begin, const It& end) const
        {
            if(begin != end) val = *begin;
        }
    };
    template<typename T>
    assign_actor_1<T> assign_a(T& v) { return assign_actor_1<T> (v); }

    /////////////////////////////////

    template<typename T,typename S>
    class assign_actor_2
    {
        T& t;
        S& s;
    public:
        explicit assign_actor_2(T&tt,S&ss) : t(tt),s(ss) { }
    
        template<typename It>
        void operator() (const It&, const It&) const
        {
            t=s;
        }
    };
    template<typename T, typename S>
    assign_actor_2<T,S> assign_a(T&t, S&s) { return assign_actor_2<T,S> (t,s); }

    /////////////////////////////////

    class save_stmt_actor
    {
        stmtp& ext;
    public:
        explicit save_stmt_actor(stmtp& e) : ext(e) { }
    
        template<typename It>
        void operator() (const It& , const It& ) const
        {
            ext = ctx.stm_result;
        }
    };
    save_stmt_actor stmtsave_a(stmtp& ext) { return save_stmt_actor(ext); }
    
    /////////////////////////////////

    class save_expr_actor
    {
        exprp& ext;
    public:
        explicit save_expr_actor(exprp& e) : ext(e) { }
    
        template<typename It>
        void operator() (const It& , const It& ) const
        {
            ext = ctx.exp_result;
        }
    };
    save_expr_actor exprsave_a(exprp& ext) { return save_expr_actor(ext); }
    
    /////////////////////////////////

    void CounterReset() { ctx.count=0; }
    void CounterInc() { ++ctx.count; }
    
    /////////////////////////////////

    void ClearVarConstruct(AP)
    {
        ctx.var_lhs = 0;
        ctx.var_result = 0;
    }
    void CatchVarName(AP)
    {
        ctx.var_result = new expr_var(yylval.str);
    }
    void AddVarIndex(AP)
    {
        ctx.var_result->addindex(ctx.exp_result);
    }
    void AddVarMember(AP)
    {
        if(ctx.var_lhs)
        {
            /* It's a member. Therefore this whole construct
             * is an user type. User types cannot contain type suffixes.
             */
            ctx.var_lhs->addmember(ctx.var_result);
        }
        else
        {
            ctx.var_lhs = ctx.var_result;
            ctx.var_result = 0;
        }
    }
    void VarFinish(AP)
    {
        SetExpr(&*ctx.var_lhs);
        ctx.var_result = ctx.var_lhs;
    }

    void ResolveAmbiguousRval(AP)
    {
        /* If it has indexes, it can be ARRAY/FUNCTION name.
         * If it has no indexes, it can be CONST/SCALAR/FUNCTIONNAME.
         * If it has members, it was an user type.
         */
        expr_var* var = ctx.var_result;
        if(!var) return;
        
        bool can_be_scalar = true;
        bool can_be_array  = true;
        bool can_be_funct  = true;
        bool can_be_const  = true;
        bool can_be_utype  = true;
        
        if(var->hasmembers())
        {
            can_be_const  = can_be_funct = false;
            can_be_scalar = can_be_array = false;
            can_be_utype  = true;
        }
        else if(var->hasindexes())
        {
            can_be_const = false;
            can_be_scalar = false;
        }
        else
        {
            can_be_array = false;
        }
    }

    /////////////////////////////////
    
    void DoBeginIf(AP)
    {
        stmt_if* i = new stmt_if;
        // TODO: type check
        i->condition = ctx.param1;
        i->thenp     = ctx.stmt1;
        ctx.stm_lhs = i;
        ctx.stmt1 = 0;
    }
    void ElseIfPivot(AP)
    {
        stmt_if* i = dynamic_cast<stmt_if*> (&*ctx.stm_lhs);
        // TODO: type check
        i->elseifp.push_back(std::make_pair(ctx.param1, ctx.stmt1));
        ctx.stmt1 = 0;
    }
    void DoEndIf(AP)
    {
        stmt_if* i = dynamic_cast<stmt_if*> (&*ctx.stm_lhs);
        i->elsep = ctx.stmt1;
        ctx.stm_result = ctx.stm_lhs;
    }

    template<typename T>
    void DoUnaryStmt(AP)
    {
        ctx.stm_result = new T(ctx.exp_result);
    }
    template<unsigned Nparams>
    void DoRestore(AP)
    {
        unsigned dataptr = 0;
        if(Nparams == 1)
        {
            expr_label* i = dynamic_cast<expr_label*> (&*ctx.exp_result);
            dataptr = TranslateLabelToDataPtr(i->get_label());
        }
        ctx.stm_result = new stmt_restore(dataptr);
    }
    
    /////////////////////////////////
    
    void DoNumericLabelRef(AP)
    {
        SetExpr(new expr_label(yylval.number));
    }
    void DoIdLabelRef(AP)
    {
        std::string id = yylval.str;
        if(HasTypeSuffix(id)) throw InvalidateAction("Not a label");
        SetExpr(new expr_label(DeclareAsLabelInCurrentScope(id)));
    }

    /////////////////////////////////
    
    void DefineGlobalDataStr(AP)
    {
        fprintf(stderr, "Data(%s)\n", yylval.str.c_str());
        global_data.push_back(yylval.str);
    }
    void DefineEmptyDataStr(AP)
    {
        fprintf(stderr, "Data(%s)\n", "");
        global_data.push_back(std::string() );
    }
    
};
#undef AP

struct Ctx1 /* enter context */
{
    void operator() () const
    {
        saved_contexts.push_back(ctx);
        ctx = Context();
    }
};
struct Ctx0 /* exit context */
{
    void operator() () const
    {
        exprp e = ctx.exp_result;
        stmtp s = ctx.stm_result;
        ctx = saved_contexts.back();
        saved_contexts.pop_back();
        
        /* If the subcontext produced an expression or a statement,
         * import them to the calling context
         */
        if(e)ctx.exp_result = e;
        if(s)ctx.stm_result = s;
    }
};

typedef rule<scanner_t, Ctx1, Ctx0> rule_t;

class QB_parser: public parser<QB_parser>
{
    rule_t program;
    rule_t sub_params_decl;
    rule_t separator, type;
    rule_t anylabeltarget, printspec;
    rule_t print_cursor_ctrl;
    rule_t plain_coordinate;
    rule_t step_coordinate;

    rule_t comparison_operator;
    
    /* The reason why there is a "longstatement" but not "longstatements0",
     * and "shortstatements0" but not "shortstatement", is as follows:
     *
     * A "longstatement" is always terminated. Either by a newline or a colon.
     * This is okay.
     *
     * However, a "shortstatement" can only be terminated by a colon. It cannot
     * be terminated by a newline.
     * A "shortstatement" does not necessarily end with a colon. For example,
     * consider this: IF a=5 THEN PRINT
     * "PRINT" is a "shortstatement" here. It does not need a colon.
     * However, the terminator cannot be made optional either.
     * PRINT PRINT is not valid, hence "*shortstatement" is not a valid
     * substitute for "shortstatements0".
     *
     */   
    
    rule_t longstatement;
    rule_t shortstatements0;
    rule_t shortif, shortif_meat;
    rule_t casematch2, casematch, dim_params, simplestatement;

    rule_t expr_IMP, expr_EQV, expr_XOR,
           expr_OR, expr_NOT, expr_AND,
           expr_Comp,
           expr_Sum, expr_MOD, expr_IDIV,
           expr_Mul, expr_Neg,
           expr_meat,
           var_target,
           autofilenum,
           ambiguous_id;

    rule_t expr;
    
    #define num_expr  expr
    #define bool_expr expr
    #define string_expr expr
    #define string_or_num_expr expr
    #define dummy_expr expr
    
public:
    QB_parser()
    {
        program =
        *(longstatement.expected()
         ) >>
        *(  switch_p
            [
                case_p<SUB> (
                    tq(IDENTIFIER) >> !sub_params_decl
                 >> !tk(STATIC) >> tk(CRLF)
                 >> *longstatement
                 >> tq(ENDSUB) >> tq(CRLF)),
                
                case_p<FUNCTION> (
                    tq(IDENTIFIER) >> !sub_params_decl
                 >> !tk(STATIC) >> separator
                 >> *longstatement
                 >> tq(ENDFUNCTION) >> tq(CRLF)),
                
                case_p<CRLF> (eps_p),
                case_p<COLONCHAR> (eps_p)
            ]
        );

        sub_params_decl =
            '('
              >> (
                   (!(tk(BYVAL) | tk(SEG))
                 >> tq(IDENTIFIER).expected()
                 >> if_p(tk(AS)) [ type ]
                    /* if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; } */
                   ) % ','
                 )
         >> ')';
        
        separator = switch_p [ case_p<COLONCHAR> (eps_p),
                               case_p<CRLF> (eps_p)
                             ];

        type = switch_p
            [  case_p<INTEGER> (eps_p),
               case_p<SINGLE> (eps_p),
               case_p<DOUBLE> (eps_p),
               case_p<STRING> (  tk('*') >> NUMBER
                               | eps_p
                              ),
               case_p<LONG> (eps_p),
               case_p<IDENTIFIER> (eps_p)
                /*
                    if(!IsType(*$1)) { YYABORT; }
                */
            ].expected();

        /**** LVAL */
        var_target =
            ambiguous_id
        |   FNIDENTIFIER;
        
        autofilenum =
            tk(FILENUM) [&DoNumber]
        |   num_expr;
        
        anylabeltarget =
            switch_p
            [
                case_p<NUMBER>     (eps_p[ &DoNumericLabelRef ]),
                case_p<IDENTIFIER> (eps_p[ &DoIdLabelRef ])
            ];
            
        printspec =
        (
            tk(USING) >> string_expr.expected() >> tq(';')
            >> (expr % (tk(',') | tk(';')))
               /* params for format, delimiter identity=insignifncant */
            >> !print_cursor_ctrl /* a print_cursor_ctrl may be given last */
        )
        | do_p[ expr | eps_p ].while_p(print_cursor_ctrl);
                /* two exprs in row is invalid, must be delimited by cursor_ctrl */
                /* many cursor_ctrls in row is valid */
                /* whether the call ends with cursor_ctrl, is important */
        
        print_cursor_ctrl = tk(',') | tk(';');

        plain_coordinate = '(' >> num_expr >> tk(',') >> num_expr >> ')';
        step_coordinate = !tk(STEP) >> plain_coordinate;

        longstatement =
        (   /* a label statement */
            /* must be checked before simplestatement */
            switch_p
            [
                case_p<NUMBER> (eps_p),
                case_p<IDENTIFIER> ( tk(COLONCHAR) )
            ]
        | ( switch_p
            [
                case_p<IF> 
                (     bool_expr [ exprsave_a(ctx.param1) ]
                   >> tk(THEN) >> tk(CRLF)
                   >> (*longstatement) [ stmtsave_a(ctx.stmt1) ]
                            [ &DoBeginIf ]
                   >> while_p(tk(ELSEIF))
                        [ bool_expr [ exprsave_a(ctx.param1) ]
                           >> tq(THEN)
                           >> (*longstatement) [ stmtsave_a(ctx.stmt1) ]
                            [ &ElseIfPivot ]
                        ]
                   >> if_p(tk(ELSE)) [ (*longstatement) [ stmtsave_a(ctx.stmt1) ] ]
                   >> tq(ENDIF)
                            [ &DoEndIf ]
                ),
                case_p<FOR>
                (     var_target >> tq('=') >> num_expr >> tq(TO) >> num_expr
                   >> if_p(tk(STEP)) [ num_expr ]
                             .else_p [ eps_p ]
                   >> separator
                   >> *longstatement
                   >> tq(NEXT)
                ),
                case_p<SELECT>
                (     tq(CASE)
                   >> expr.expected() >> +separator
                   >> while_p(tk(CASE))
                       [ casematch2 >> +separator >> *longstatement ]
                   >> tq(ENDSELECT)
                ),
                 /* do while ... loop
                    do until ... loop
                    do ... loop
                    do ... loop while
                    do ... loop until */
                case_p<DO> (
                      separator
                   >> *longstatement
                   >> switch_p
                      [
                        case_p<LOOP> (eps_p),
                        case_p<LOOPWHILE> (bool_expr),
                        case_p<LOOPUNTIL> (bool_expr)
                      ].expected()),
                case_p<DOWHILE> (
                      bool_expr >> separator
                   >> *longstatement
                   >> tq(LOOP)),
                case_p<DOUNTIL> (
                      bool_expr >> separator
                   >> *longstatement
                   >> tq(LOOP)),
                case_p<DEF> (
                    /* multiline DEF. single-line DEF is included in simplestatement. */
                    tq(FNIDENTIFIER) >> !sub_params_decl >> tk(CRLF)
                    >> *longstatement
                    >> tq(ENDDEF)
                  ),
                /* TODO: TYPE cannot exist in SUBs.
                 * However, Quickbasic allows them inside control structures.
                 * But they are unconditional.
                 */
                case_p<TYPE> (
                    tq(IDENTIFIER) >> tq(CRLF)
                 >>(+(
                      *tk(CRLF)/* empty line is ok in typedef */
                    >>
                     (
                      (
                       (tk(IDENTIFIER) >> tq(AS) >> type
                            /* { define it
                               }
                           */
                       ) % ','
                      )
                      >> tq(CRLF)
                     )
                   ) ).expected()
                 >> tq(ENDTYPE)
                )
            ]
          | simplestatement
          ) >> separator.expected()
        |   shortif >> tq(CRLF)
        |   separator
        );
        
        /* shortstatements0 does not include trailing COLONCHARs. */
        shortstatements0 =
            *tk(COLONCHAR)
        >>  /* statements that can be included between online THEN and ELSE */
            /* Does not include shortif. */
          !(
            (   switch_p
                [
                    case_p<FOR> (
                          var_target >> tq('=') >> num_expr >> tq(TO) >> num_expr
                       >> if_p(tk(STEP)) [ num_expr ]
                                 .else_p [ eps_p ]
                       >> tk(COLONCHAR)
                       >> shortstatements0 >> *tk(COLONCHAR)
                       >> tq(NEXT)),
                    case_p<SELECT> (
                          tq(CASE) >> expr.expected()
                       >> +tk(COLONCHAR)
                       >> while_p(tk(CASE))
                            [ casematch2 >> +tk(COLONCHAR)
                           >> shortstatements0 >> +tk(COLONCHAR) ]
                       >> tq(ENDSELECT)),
                    case_p<DO> (
                          tq(COLONCHAR)
                       >> shortstatements0 >> *tk(COLONCHAR)
                       >> switch_p
                          [
                            case_p<LOOP> (eps_p),
                            case_p<LOOPWHILE> (bool_expr),
                            case_p<LOOPUNTIL> (bool_expr)
                          ].expected()),
                    case_p<DOWHILE> (
                          bool_expr >> tk(COLONCHAR)
                       >> shortstatements0 >> *tk(COLONCHAR)
                       >> tq(LOOP)),
                    case_p<DOUNTIL> (
                          bool_expr >> tk(COLONCHAR)
                       >> shortstatements0 >> *tk(COLONCHAR)
                       >> tq(LOOP))
                ]
            |   simplestatement
            ) % +tk(COLONCHAR)
           );
        
        shortif =
            tk(IF) >> bool_expr [ exprsave_a(ctx.param1) ]
            >> tk(THEN) >> shortif_meat [ stmtsave_a(ctx.stmt1) ]
                 [ &DoBeginIf ]
            >> *tk(COLONCHAR)
            >> (if_p(tk(ELSE)) [ shortif_meat [ stmtsave_a(ctx.stmt1) ] ])
                 [ &DoEndIf ] 
            >> *tk(COLONCHAR);
        
        shortif_meat =
            tk(NUMBER) [ &DoNumericLabelRef ]
                [ &DoUnaryStmt<stmt_goto> ] /* Translate into goto */
        |   shortif
        |   shortstatements0
            >> !(+tk(COLONCHAR) >> shortif);
        
        comparison_operator =
          switch_p
          [
            case_p<NOTEQUAL> (eps_p),
            case_p<LESSOREQUAL> (eps_p),
            case_p<GREATEROREQUAL> (eps_p),
            case_p<'>'> (eps_p),
            case_p<'<'> (eps_p),
            case_p<'='> (eps_p)
          ];
        
        casematch2 = casematch
         >> while_p( tk(',') )
                [ casematch [ int_act_a<DoBinaryExpr<expr_caseor> >(0) ] ];
        
        casematch =
                tk(ELSE)                    /* $$ = new casevalues(ELSE);   */
                    [ int_act_a<DoZnaryExpr<expr_caseelse> >(0) ]
            |   tk(IS) >> comparison_operator.expected() [ assign_a(ctx.token) ]
                       >> expr
                    [ int_act_ref_a<DoUnaryExpr<expr_caseis> >(ctx.token) ]
            |	expr [ exprsave_a(ctx.param1) ]
                     >> ( tk(TO) >> expr.expected() [ exprsave_a(ctx.param2) ]
                    [ int_act_a<Do_2_ParamFunc<expr_caserange> > (0) ]
                        | eps_p
                    [ int_act_a<DoUnaryExpr<expr_caseis> > ('=') ]
                        );

        dim_params =
            /* DIM and REDIM.
             * Note: REDIM can only be applied to $DYNAMIC arrays,
             *             not scalars, utypes or $STATIC arrays.
             */
            !tk(SHARED) /* optional word */
            >>  ((
                    tk(IDENTIFIER)
                    /* if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; } */
                 >> if_p(tk('(')) /* array dimension */
                    [ (( num_expr >> if_p(tk(TO)) [ num_expr ]
                       ) % ',') >> ')' ]
                 >> if_p(tk(AS)) [ type ]
                 ) % ',');
        
        simplestatement =
            /* This switch_p is split into multiple, to shorten the compilation time */
            switch_p
            [
                /* */
                case_p<LET> (
                    var_target >> tq('=') >> expr
                    ),
                
                /* explicit sub call */
                case_p<CALL> (
                    tk(IDENTIFIER) >> !('(' >> (expr % ',') >> ')')
                    /*
                        if(IsDefinedAsNonSub(*$2)) { YYABORT; }
                    */
                    ),
                
                /* single-line DEF. multiline DEF is already caught in longstatement. */
                /* TODO: Don't allow inside a control structure such as IF,FOR,SUB. */
                case_p<DEF> (
                    tq(FNIDENTIFIER) >> !sub_params_decl
                    >> tq('=') >> expr
                ),
               
                /* Note: QuickBasic allows CONSTs in control structures.
                 * But they are unconditional.
                 */
                case_p<CONST> (
                    tq(IDENTIFIER) >> tq('=') >> expr.expected()
                ),
                
                /* TODO: OPTION BASE cannot exist in SUBs.
                 * Note: QuickBasic allows OPTION BASE in control structures.
                 * But they are unconditional. Also, only 1 is allowed.
                 * Attempt to use more will produce compile-time error
                 * "Array already dimensioned".
                 */
                case_p<OPTION_BASE> (
                    tq(NUMBER) [&DoNumber] [exprsave_a(option_base)]
                ),

                /* TODO: DATA cannot exist in SUBs.
                 * However, Quickbasic allows them inside control structures.
                 * But they are unconditional.
                 */
                case_p<DATA> (
                   ((tk(STRINGCONST) [ &DefineGlobalDataStr ]
                   | eps_p           [ &DefineEmptyDataStr ]
                    ) % ',')
                ),
                
                /* TODO: STATIC keyword is invalid outside SUB/FUNCTION. */
                case_p<STATIC> (
                    (
                       tk(IDENTIFIER)
                    >> if_p(tk('(')) [ tk(')') ]
                    >> if_p(tk(AS)) [ type ]
                    ) % ','
                ),

                /* TODO: STATIC keyword is invalid outside SUB/FUNCTION. */
                case_p<SHARED> (
                    (
                       tk(IDENTIFIER)
                    >> if_p(tk('(')) [ tk(')') ]
                    >> if_p(tk(AS)) [ type ]
                    ) % ','
                ),

                /* TODO: DECLARESUB is invalid after the first statement and in SUB/FUNC */
                case_p<DECLARESUB> (
                    tq(IDENTIFIER) >> ((tk('(') >> ')') | sub_params_decl )
                 >> !tk(CDECL)
                 >> if_p(tk(ALIAS)) [tk(STRINGCONST)]
                ),
                
                /* TODO: DECLAREFUNCTION is invalid after the first statement and in SUB/FUNC */
                case_p<DECLAREFUNCTION> (
                    tq(IDENTIFIER) >> ((tk('(') >> ')') | sub_params_decl )
                 >> !tk(CDECL)
                 >> if_p(tk(ALIAS)) [tk(STRINGCONST)]
                ),
                 
                /* TODO: COMMON is invalid after the first statement and in SUB/FUNC */
                case_p<COMMON> (
                   !tk(SHARED)
                 >> if_p(tk('/')) [ tk(IDENTIFIER) >> '/' ]
                 >> (tq(IDENTIFIER) % ',')
                 >> tq(CRLF)
                )
            ]
            
          | switch_p
            [
                /* data commands */
                
                case_p<RESTORE>
                (
                    anylabeltarget [ &DoRestore<1> ]
                  | eps_p [ &DoRestore<0> ]
                ),
                case_p<READ>
                (
                    var_target [ &DoUnaryStmt<stmt_read> ]
                    % ','
                ),
               
                case_p<DIM> (dim_params),
                case_p<REDIM> (dim_params),
                case_p<ERASE> (
                        /* For static arrays, ERASE clears their contents to zero.
                         * For dynamic arrays, ERASE shrinks them to zero. */
                        var_target % ','
                    ),
                
                /* system commands */
                case_p<SYSTEM>(
                    !string_expr),
                case_p<POKE>(
                    num_expr >> tk(',') >> num_expr),
                case_p<OUT>(
                    num_expr >> tk(',') >> num_expr),
                case_p<WAIT>(
                    num_expr
                 >> tk(',')
                 >> num_expr
                 >> if_p(tk(',')) [ num_expr ] ),
                case_p<DEFSEG>(
                    if_p(tk('=')) [ num_expr ] ),
                
                /* control commands */
                
                case_p<EXITDEF> ( eps_p ),
                case_p<EXITDO>( eps_p ),
                case_p<EXITFOR>( eps_p ),
                case_p<EXITSUB>( eps_p ),
                case_p<EXITFUNCTION>( eps_p ),
                case_p<STOP>( eps_p ),
                case_p<END>( eps_p ),
                case_p<RESUME>(
                    tk(NEXT)
                |   anylabeltarget
                |   eps_p),
                case_p<RETURN>(
                    !anylabeltarget),
                case_p<GOTO>(
                    anylabeltarget),
                case_p<GOSUB>(
                    anylabeltarget)
            ]
          | switch_p
            [
                /* graphics commands */
                
                case_p<LOCATE>(
                    !num_expr >> repeat_p(0,4) [tk(',') >> !num_expr ]
                    /* row, col, cursor_vis, cursor_start, cursor_stop */
                    ), /* cursor_end */
                case_p<SCREEN>(
                    !num_expr >> repeat_p(0,3) [tk(',') >> !num_expr ]
                    /* mode, colorswitch, apage, vpage */
                    ),
                case_p<PAINT>(
                       step_coordinate
                    >> !(tk(',')
                    >> !string_or_num_expr /* pattern/paint */
                    >> !(tk(',')
                    >> !num_expr /* bordercolor */
                    >> !(tk(',')
                    >> !string_expr)))), /* background */
                case_p<COLOR>(
                    !num_expr >> repeat_p(0,2) [tk(',') >> !num_expr ]
                    /* fgcolor, bgcolor, border */
                ),
                case_p<CIRCLE>(
                         step_coordinate
                      >> tk(',') >> num_expr /* radius */
                      >> repeat_p(0,4) [tk(',') >> !num_expr]
                        /* color, start, end, aspect */
                ),
                case_p<GET>(
                       step_coordinate >> tk('-') >> step_coordinate
                    >> tk(',') >> var_target),
                case_p<PUT>(
                       step_coordinate
                    >> tk(',') >> var_target
                    >> if_p(tk(','))
                        [ tk(XOR) | tk(PSET) | tk(PRESET) | tk(AND) | tk(OR) ]),
                        /* tk(XOR) is default */
                case_p<VIEWPRINT>(
                        !(num_expr >> tk(TO) >> num_expr)),
                case_p<VIEW>(
                        !tk(SCREEN)
                     >> plain_coordinate >> tk('-') >> plain_coordinate
                     >> !(tk(',')
                     >> !num_expr /* color */
                     >> !(tk(',')
                     >> !num_expr))), /* border */
                case_p<PALETTE>(
                        tk(USING) >> var_target
                    |   !(num_expr >> tk(',') >> num_expr)),
                case_p<PSET>(
                        step_coordinate
                     >> !(tk(',')
                     >> !num_expr)), /* color */
                case_p<PRESET>(
                        step_coordinate
                       >> !(tk(',')
                       >> !num_expr)), /* color */
                case_p<LINE>(
                        !step_coordinate >> tk('-') >> step_coordinate
                     >> !(tk(',')
                     >> !num_expr /* color */
                     >> !(tk(',')
                     >> !tk(IDENTIFIER) /* tk(B) | tk(BF) */
                     >> !(tk(',')
                     >> !num_expr)))), /* style (16-bit int mask) */
                case_p<DRAW>(
                        string_expr),
                case_p<CLS>(
                        !num_expr)
            ]
          | switch_p
            [
                /* I/O commands */
                
                case_p<OPEN>(
                    string_expr /* filename */
                     >> if_p(tk(FOR))
                         [ tk(RANDOM) | tk(OUTPUT) | tk(INPUT) | tk(APPEND) | tk(BINARY) ]
                         /* RANDOM is default */
                     >> if_p(tk(ACCESS))
                         [ tk(READ) || tk(WRITE) ]
                     >> !(tk(SHARED)
                       |  tk(LOCK) >> (tk(READ) || tk(WRITE)))
                     >> tk(AS) >> autofilenum
                     >> if_p(tk(LEN)) [ tq('=') >> num_expr ]
                |   string_expr /* mode, first char is O/I/R/B/A */
                     >> tk(',') >>autofilenum
                     >> string_expr /* filename */
                     >> if_p(tk(',')) [ num_expr ]), /* reclen */
                case_p<CLOSE>(
                    !(autofilenum % ',')), /* empty list = close all files */
                case_p<PRINT>(
                    printspec),
                case_p<PRINT_F>(
                    autofilenum >> tk(',') >> printspec),
                case_p<LPRINT>(
                    printspec),
                
                /* event commands */
                case_p<PLAY> (
                    (switch_p
                     [
                        case_p<STOP> ( eps_p ),
                        case_p<OFF>  ( eps_p ),
                        case_p<ON>   ( eps_p )
                     ]
                   | (string_expr % ',')
                    ).expected()
                     ),
                case_p<TIMER> (
                     switch_p
                     [
                        case_p<STOP> ( eps_p ),
                        case_p<OFF>  ( eps_p ),
                        case_p<ON>   ( eps_p )
                     ].expected()),
                case_p<KEY> (
                    (switch_p
                     [
                        case_p<LIST> ( eps_p ),
                        case_p<OFF>  ( eps_p ),
                        case_p<ON>   ( eps_p )
                     ]
                   |  num_expr >> tk(',') >> string_expr
                    ).expected()
                     ),
                case_p<ON_PLAY> (
                    '(' >> num_expr >> ')' >> tk(GOSUB) >> anylabeltarget
                        ),
                case_p<ON_KEY> (
                    '(' >> num_expr >> ')' >> tk(GOSUB) >> anylabeltarget
                        ),
                case_p<ON_TIMER> (
                    '(' >> num_expr >> ')' >> tk(GOSUB) >> anylabeltarget
                        ),
                case_p<ON_ERROR> (
                    tk(GOTO) >> anylabeltarget
                        ),
                case_p<ON> (
                    /* translate these as SELECT CASE CINT(expr) */
                    
                    num_expr >> tk(GOSUB) >> (anylabeltarget % ',')
                |   num_expr >> tk(GOTO)  >> (anylabeltarget % ',')
                    )
            ]
        
        |   /* assign */
            var_target >> '=' >> expr
            
        |   /* implicit sub call */
            /* Must be tested after assignstatement */
            tk(IDENTIFIER) >> !(expr % ',')  /* Calling a SUB */
            /*
                if(IsDefinedAsNonSub(*$1)) { YYABORT; }
            */
            ;
        
        /* FIXME: Check right context */
        #define binary_op_while(subrule, op) \
            while_p( (op) [ assign_a(ctx.token) ] ) \
                [ (subrule) >> eps_p \
                   [ int_act_ref_a<DoBinaryExpr<expr_binary> >(ctx.token) ] \
                ]
    
        #define unary_toggleop_while(subrule, op, type) \
            (for_p(&CounterReset, (op), &CounterInc) [ eps_p ] \
            >> subrule) \
              [ int_act_ref_a<DoUnaryToggleExpr<expr_unary, type> >(ctx.count) ]
        
        expr     = expr_IMP >> binary_op_while(expr_IMP, tk(IMP) ); // a IMP b = b OR NOT a
        expr_IMP = expr_EQV >> binary_op_while(expr_EQV, tk(EQV) ); // a EQV b = NOT(a XOR b)
        expr_EQV = expr_XOR >> binary_op_while(expr_XOR, tk(XOR) );
        expr_XOR = expr_OR  >> binary_op_while(expr_OR , tk(OR ) );
        expr_OR  = expr_AND >> binary_op_while(expr_AND, tk(AND) );
        
        expr_AND = unary_toggleop_while(expr_NOT, tk(NOT), NOT);
        
        expr_NOT  = expr_Comp >> binary_op_while(expr_Comp, comparison_operator);
        expr_Comp = expr_Sum  >> binary_op_while(expr_Sum,  (switch_p[case_p<'+'>(eps_p),case_p<'-'>(eps_p)]) );
        expr_Sum  = expr_MOD  >> binary_op_while(expr_MOD,  tk(MOD) );
        expr_MOD  = expr_IDIV >> binary_op_while(expr_IDIV, tk('\\') );
        expr_IDIV = expr_Mul  >> binary_op_while(expr_Mul,  (switch_p[case_p<'*'>(eps_p),case_p<'/'>(eps_p)]) );

        expr_Mul = unary_toggleop_while(expr_Neg, tk('-'), '-');

        expr_Neg = expr_meat >> binary_op_while(expr_Mul, tk('^') );
        
        expr_meat =
            '(' >> expr >> ')'
            
          #define Do_0_Param(ret_type) \
            [ assign_a(ctx.token) ] \
                [ int_act_ref_a<DoZnaryExpr<ret_type> >(ctx.token) ]

          #define Do_1_Param(ret_type, p1type) \
            [ assign_a(ctx.token) ] \
             >> ('(' >> p1type >> ')') \
                [ int_act_ref_a<DoUnaryExpr<ret_type> >(ctx.token) ]
          
          #define Do_2_Param(ret_type, p1type,p2type) \
            [ assign_a(ctx.token) ] \
             >> ('(' >> p1type [exprsave_a(ctx.param1)] \
                     >> ',' \
                     >> p2type [exprsave_a(ctx.param2)] \
                     >> ')') \
                [ int_act_ref_a<Do_2_ParamFunc<ret_type> >(ctx.token) ]
           
          #define Do_1or2_Param(ret_type, p1type,p2type) \
            [ assign_a(ctx.token) ] \
             >> '(' >> p1type [exprsave_a(ctx.param1)] >> \
              (        tk(')') \
                [ int_act_ref_a<DoUnaryExpr<ret_type> >(ctx.token) ] \
                     | p2type [exprsave_a(ctx.param2)] >> \
                       tk(')') \
                [ int_act_ref_a<Do_2_ParamFunc<ret_type> >(ctx.token) ] \
              )
        
      #if 1
        #define a switch_p[case_p<
        #define i >(eps_p),case_p<
        #define e >(eps_p)]
      #else
        #define a (ch_p(
        #define i ) | ch_p(
        #define e ))
      #endif
        
        |   // numeric -> numeric functions
            a ABS i ATN i CDBL i CINT i CLNG i COS i CSNG i EXP
            i FIX i INP i INT i LOG i LPOS i PEEK i SPC i SQR
            i TAB i TAN e
             Do_1_Param(expr_builtin_numeric, num_expr)
            
            // string -> numeric functions
          | a ASC i CVD i CVI i CVL i CVS i VAL e
             Do_1_Param(expr_builtin_numeric, string_expr)
           
            // numeric -> string functions
          | a CHR_S i HEX_S i MKD_S i MKI_S i MKL_S
            i MKS_S i OCT_S i SPACE_S i STR_S e
             Do_1_Param(expr_builtin_string, num_expr)

            // string -> string functions
          | a LCASE_S i LTRIM_S i RTRIM_S i UCASE_S e
             Do_1_Param(expr_builtin_string, string_expr)

            // implied -> numeric functions
          | a CSRLIN i ERL i ERR i FREEFILE i TIMER e
             Do_0_Param(expr_builtin_numeric)

            // implied -> string functions
          | a DATE_S i INKEY_S i TIME_S e
             Do_0_Param(expr_builtin_string)
           
            // autofilenum -> string
          | a LOC i LOF e
             Do_1_Param(expr_builtin_numeric, autofilenum)
          
          | tk(ENVIRON_S)
             Do_1_Param(expr_builtin_string, string_or_num_expr)
          
          | tk(LEN)
             Do_1_Param(expr_builtin_numeric, expr) /* any type */

          | tk(FRE)
             Do_1_Param(expr_builtin_numeric, string_or_num_expr)
          
          | a PLAY i POS e
             Do_1_Param(expr_builtin_numeric, dummy_expr)
          
          | tk(FILEATTR)
             Do_2_Param(expr_builtin_numeric, autofilenum, num_expr)
          
          | a LEFT_S i RIGHT_S e
             Do_2_Param(expr_builtin_string, string_expr, num_expr)
          
          | tk(STRING_S)
             Do_2_Param(expr_builtin_string, num_expr, string_or_num_expr)
          
          | a VARPTR i VARSEG e
             Do_1_Param(expr_builtin_numeric, var_target)
          
          | tk(VARPTR_S)
             Do_1_Param(expr_builtin_string, var_target)
          
          | tk(INPUT_S)
             Do_1or2_Param(expr_builtin_string, num_expr, autofilenum)
          
          | tk(POINT)
             Do_1or2_Param(expr_builtin_numeric, num_expr, num_expr)
          
          | tk(LBOUND)
             Do_1or2_Param(expr_lbound, var_target, num_expr)

          | tk(UBOUND)
             Do_1or2_Param(expr_ubound, var_target, num_expr)

            // non-general type functions
          | switch_p
            [
                case_p<INSTR>(
                    '(' >> string_expr [exprsave_a(ctx.param1)]
                        >> tk(',')
                        >> string_expr [exprsave_a(ctx.param2)]
                        >> tk(')')
                    [ int_act_a<Do_2_ParamFunc<expr_builtin_string> >(INSTR) ]
                   |'(' >> num_expr [exprsave_a(ctx.param1)]
                        >> ','
                        >> string_expr [exprsave_a(ctx.param2)]
                        >> ','
                        >> string_expr [exprsave_a(ctx.param3)]
                        >> tk(')')
                    [ int_act_a<Do_3_ParamFunc<expr_builtin_string> >(INSTR) ]
                    ),
                case_p<MID_S>(
                    '(' >> string_expr [exprsave_a(ctx.param1)]
                        >> ','
                        >> num_expr [exprsave_a(ctx.param2)]
                        >> (
                            tk(')')
                    [ int_act_a<Do_2_ParamFunc<expr_builtin_string> >(MID_S) ]
                          | (',' >> num_expr [exprsave_a(ctx.param3)] >> ')')
                    [ int_act_a<Do_3_ParamFunc<expr_builtin_string> >(MID_S) ]
                           )
                    ),
                case_p<RND>(
                         ('(' >> num_expr >> ')')
                    [ int_act_a<DoUnaryExpr<expr_builtin_numeric> >(RND) ]
                        | eps_p
                    [ int_act_a<DoZnaryExpr<expr_builtin_numeric> >(RND) ]
                    ),
                
                case_p<NUMBER>(      eps_p [ &DoNumber ] ),
                case_p<STRINGCONST>( eps_p [ &DoString ] ),
                case_p<FNIDENTIFIER>(
                    '(' >> (expr % ',') >> ')'
                  | eps_p
                )
            ]
        
        #undef a
        #undef i
        #undef e
        
        |   /**** RVAL */
            /* Var or const or function call.
             * Collect it into an expr_var first. Then decide if to convert
             * it to expr_userfun.
             */
            ambiguous_id [ &ResolveAmbiguousRval ];

    ambiguous_id =
            eps_p
                [ &ClearVarConstruct ]
         >> ((
                tk(IDENTIFIER) [ &CatchVarName ]
                /* perhaps with indexes */
                >> if_p(tk('('))
                   [
                      (num_expr [ &AddVarIndex ]
                        % ','
                      )
                        >> ')'
                   ]
            ) [ &AddVarMember ]
            % '.') [ &VarFinish] ;
    

program.setname("program");
sub_params_decl.setname("sub_params_decl");
separator.setname("separator");
type.setname("type");
anylabeltarget.setname("anylabeltarget");
printspec.setname("printspec");
print_cursor_ctrl.setname("print_cursor_ctrl");
plain_coordinate.setname("plain_coordinate");
step_coordinate.setname("step_coordinate");
comparison_operator.setname("comparison_operator");
longstatement.setname("longstatement");
shortstatements0.setname("shortstatements0");
shortif.setname("shortif");
shortif_meat.setname("shortif_meat");
casematch2.setname("casematch2");
casematch.setname("casematch");
dim_params.setname("dim_params");
simplestatement.setname("simplestatement");
expr_IMP.setname("expr_IMP");
expr_EQV.setname("expr_EQV");
expr_XOR.setname("expr_XOR");
expr_OR.setname("expr_OR");
expr_NOT.setname("expr_NOT");
expr_AND.setname("expr_AND");
expr_Comp.setname("expr_Comp");
expr_Sum.setname("expr_Sum");
expr_MOD.setname("expr_MOD");
expr_IDIV.setname("expr_IDIV");
expr_Mul.setname("expr_Mul");
expr_Neg.setname("expr_Neg");
expr_meat.setname("expr_meat");
var_target.setname("var_target");
autofilenum.setname("autofilenum");
ambiguous_id.setname("ambiguous_id");
expr.setname("expr");
    }

    bool parse(scanner_t& scanner) const
    {
        return program.parse(scanner);
    }
    
    #undef expr
};

/************************************/

int main(void)
{
    /*setbuf(stdin,NULL);*/
    
    yyStartFile(stdin, "<stdin>");
    
    Lex<QBtoken> lexer;
    scanner_t sca(lexer.begin(), lexer.end());
    QB_parser pa;
    
    pa.parse(sca);
    
    if(!sca.eof())
        fprintf(stderr, "Parse error\n");

    return 0;
}
