template<typename ActorT>
class LbtP: public parser<LbtP<ActorT> >
{
    typedef parser<LbtP<ActorT> > base_t;
    const ActorT actor;
    const std::string name;
public:
    LbtP(const ActorT& act)
        : base_t(), actor(act), name() { }
    LbtP(const ActorT& act, const std::string& n)
        : base_t(), actor(act), name(n) { }

    template<typename ScannerT>
    bool parse(ScannerT& scan) const
    {
        if(scan.eof()) return false;
        typename ScannerT::iterator it = scan.pos();
        typename ScannerT::value_type v = scan.get();
        try
        {
            if(actor(v, scan)) return true;
        }
        catch(const InvalidateAction& inv)
        {
            yyerror(inv.what(), it.GetLoc());
        }
        scan.revert(it);
        return false;
    }

    const std::string what() const
    {
        return name.empty() ? "lambda_tok_parser" : name;
    }
};

template<typename ActorT>
const LbtP<ActorT>
LambdaTokParser(const ActorT& actor) { return LbtP<ActorT> (actor); }

template<typename ActorT>
const LbtP<ActorT>
LambdaTokParser(const ActorT& actor, const std::string& name)
    { return LbtP<ActorT> (actor, name); }

/////////////

template<typename ActorT>
class LbdP: public parser<LbdP<ActorT> >
{
    typedef parser<LbdP<ActorT> > base_t;
    const ActorT actor;
    const std::string name;
public:
    LbdP(const ActorT& act)
        : base_t(), actor(act), name() { }
    LbdP(const ActorT& act, const std::string& n)
        : base_t(), actor(act), name(n) { }

    template<typename ScannerT>
    bool parse(ScannerT& scan) const
    {
        typename ScannerT::iterator it = scan.pos();
        try
        {
            if(actor(scan)) return true;
        }
        catch(const InvalidateAction& inv)
        {
            yyerror(inv.what(), scan.eof() ? LexLocation() : it.GetLoc());
        }
        scan.revert(it);
        return false;
    }

    const std::string what() const
    {
        return name.empty() ? "lambda_parser" : name;
    }
};

template<typename ActorT>
const LbdP<ActorT>
LambdaParser(const ActorT& actor) { return LbdP<ActorT> (actor); }

template<typename ActorT>
const LbdP<ActorT>
LambdaParser(const ActorT& actor, const std::string& name)
    { return LbdP<ActorT> (actor, name); }

