#ifndef bqtQBlibErrorHH
#define bqtQBlibErrorHH

#undef B_ERROR
#define B_ERROR_INCLUDED

/* NOTE: This relies on a GNU extension. */

#include "types.hh"

#include <cstdio>
#include <vector>

struct B_ErrorEv     { };
struct B_ErrorPacket_Brief { void *repeat, *ignore; };
struct B_ErrorPacket_Full  { void *repeat, *ignore; const char* filename; unsigned line, column; };

#define B_ERR_STMT_BEGIN(Repeat_Label) \
    Repeat_Label: \
    try {

#define B_ERR_STMT_END(Repeat_Label,Ignore_Label, line, col, filename) \
    } \
    catch(const B_ErrorEv&) \
    { \
        static const B_ErrorPacket_Full err {&&Repeat_Label,&&Ignore_Label, filename,line,col }; \
        B_ERROR_Init2(err); \
        goto *B_ErrorHandler; \
        Ignore_Label:; \
    }

#define B_ERR_STMT_END_BRIEF(Repeat_Label,Ignore_Label) \
    } \
    catch(const B_ErrorEv&) \
    { \
        static const B_ErrorPacket_Brief err { &&Repeat_Label,&&Ignore_Label }; \
        B_ERROR_Init(err); \
        goto *B_ErrorHandler; \
        Ignore_Label:; \
    }

#ifdef B_NO_ERROR_HANDLING
# define B_ERR_FRAME_MAIN_BEGIN() \
        void* EventReturn = nullptr; \
        B_ErrorHandling_Enabled = false; \
        if(false) { B_Event_Dispatch: goto *B_GetEvent(EventReturn); goto B_Event_Dispatch; } \
     try {
# define B_ERR_FRAME_MAIN_END() \
     } catch(const B_ErrorEv&) \
     { \
        static const B_ErrorPacket_Brief err { &&Main_Error,&&Main_Error }; \
        B_ERROR_Init(err); \
        Main_Error:; \
     }
#else
# define B_ERR_FRAME_MAIN_BEGIN() \
        void* EventReturn = nullptr; \
        if(false) { B_Event_Dispatch: goto *B_GetEvent(EventReturn); goto B_Event_Dispatch; }
# define B_ERR_FRAME_MAIN_END()   /*nothing to do */
#endif

#define B_ERR_FRAME_SUB_BEGIN() \
        void* EventReturn = nullptr; \
        if(false) { B_Event_Dispatch: goto *B_GetEvent(EventReturn); goto B_Event_Dispatch; }

#define B_ERR_FRAME_SUB_END()   /*nothing to do */

extern "C" {

extern bool  B_ErrorHandling_Enabled;
extern void* B_ErrorHandler;
extern int   B_ERR_VALUE, B_ERL_VALUE;
void B_ERROR_Init2(const B_ErrorPacket_Full& err) __attribute__((regparm(1)));
void B_ERROR_Init(const B_ErrorPacket_Brief& err) __attribute__((regparm(1)));
void* B_RESUME_Execute(int mode) __attribute__((regparm(1)));

/* These macros are invoked by the actual program */
#define B_RESUME()         goto* B_RESUME_Execute(1)
#define B_RESUME_NEXT()    goto* B_RESUME_Execute(0)
#define B_RESUME_TO(label) do { B_RESUME_Execute(2); goto label; } while(0)

void B_StmtERROR(int code);

#ifndef B_NO_ERROR_HANDLING
  #define B_StmtERROR(code) do { B_ERR_VALUE = (code); throw B_ErrorEv{}; } while(0)
#endif

#define ON_ERROR_GOTO(label)    B_ErrorHandler = &&label
#define ON_ERROR_GOTO_NOWHERE() B_ErrorHandler = nullptr
static inline int B_FuncERR() { return B_ERR_VALUE; }
static inline int B_FuncERL() { return B_ERL_VALUE; }

/* Utilities used by library routines. */
void B_ERROR(int code) __attribute__((noreturn));
int B_ERROR_TranslateErrno();
#define B_ERROR_Errno() B_ERROR(B_ERROR_TranslateErrno())


}

#endif
