#ifndef bqtSymTablesHH
#define bqtSymTablesHH

#include <map>
#include <set>
#include <vector>
#include <string>
#include <utility>
#include <unordered_map>
#include <memory>

#include "types.hh"
#include "expr.hh"

#include "lexer.hh" // For LexLocation

struct TypeFlagSet
{
    //union {
        struct{
    bool is_subroutine:1;
    bool is_scalar    :1;
    bool is_const     :1;
    bool is_array     :1;

    bool is_global_scalar : 1; // imported using SHARED
    bool is_global_array  : 1; // imported using SHARED

    // In case of subroutines, they can be DECLARED and later defined.
    bool has_body     :1;
        };
        //unsigned zero : 8;
    //};

    // Types and labels not included here; they are in different namespace.

    TypeFlagSet() :
        is_subroutine(0), is_scalar(0),
        is_const(0), is_array(0),
        is_global_scalar(0), is_global_array(0),
        has_body(0) {}
};

/* SUB, FUNCTION, DEF FN */
struct SubRoutine
{
    struct ParamSpec
    {
        std::string name;
        TypeP       type;
        bool        is_array;
        enum { ByRef, ByVal, Seg } passconvention;
        bool        was_modified; // If the param is never modified,
                                  // It can be modified into ByVal.

        ParamSpec() : name(), type(), is_array(false),
                      passconvention(ByRef), was_modified(false) { }

        void SetByval() { passconvention=ByVal; }
        void SetSeg() { passconvention=Seg; }
    };

    LexLocation defined_at, declared_at;

    TypeP returntype = nullptr; // 0=SUB, otherwise FUNCTION/DEF
    std::vector<ParamSpec> params;
    bool is_static = false;

    std::unique_ptr<class stmt> code;

    int callconvention = 0; // 0=default, 1=cdecl
    std::string linkername; // ALIAS

    struct Scope* sub_scope = nullptr;
};

struct LabelDef
{
    bool     defined;
    unsigned dataptr;
};

struct VarContextDef
{
    LexLocation defined_at;
    TypeP type     = nullptr;
    bool is_static = false; // STATIC to the given function
    bool is_shared = false; // SHARED to the subfunctions (towards children)
    int param_index = -1;
    //stmtp first_use;
};

typedef std::unordered_map<std::string, VarContextDef> ScopeVarListType;

struct Scope
{
    /*****
      Method of lookup:
        Check constants. Ignore suffix.
          Match? Fail if had wrong suffix. Otherwise, accept.

        Check UTYPE.
          Match? Fail if had a suffix. Otherwise, accept.
        Had suffix? Check by suffix.

        Otherwise, check if in Utypes.
          If the symbol was declared using DIM...AS,
          this gets the explicit type. DEFtype is the
          last resort.

        Otherwise, check by DEFtype.
    *******/

    std::unordered_map<std::string, UtypeP> types;    // Types have no suffix.
    std::unordered_map<std::string, exprp> constants; // Constants have no suffix.

    // SUBs, FUNCTIONs, DEFs
    std::unordered_map<std::string, SubRoutine> subroutines;

    ScopeVarListType Scopes[NumTypes][2]; // indexed by TypeType. Second index: 0=scalars, 1=arrays

    std::unordered_map<std::string, TypeFlagSet> usage;

    // Labels... Original name -> translated name
    std::unordered_map<std::string, std::string> labelnames;
    std::map<std::string, LabelDef> labels;
};

#endif
