#ifndef bqtTypesHH
#define bqtTypesHH

#include <memory>
#include <string>
#include <vector>
#include <utility>
#include <unordered_map>

typedef std::shared_ptr<struct TypeInfo> TypeP;
typedef std::shared_ptr<struct UserType> UtypeP;

extern const TypeP TypeInt, TypeSng, TypeDbl, TypeLng, TypeStr;
extern const TypeP TypeByte, TypeUbyte, TypeUint, TypeUlong, TypeBit, TypeUbit;

enum class TypeType
{
    UserType,
    Single,
    Double,
    String,
    Integer,
    Long,
    Bit,
    Byte,
    UnsignedInteger,
    UnsignedLong,
    UnsignedBit,
    UnsignedByte
};
// Note: The last "Type" must be written in NumTypes.
constexpr unsigned NumTypes = 1+(int)TypeType::UnsignedByte;

struct TypeInfo
{
    TypeType Type;

    enum ArrayType
    {
        NotArray,
        StaticArray,      // '$STATIC
        DynamicArray,     // '$DYNAMIC
        ParamArray        // One passed via parameter (number of dimensions unknown)
    };

    ArrayType arraytype = NotArray;           //if array
    unsigned fixlength = 0;                   //if non-dynamic string.
    std::string utypename;                    //if UserType
    const struct UserType* members = nullptr; //if UserType, provided for convenience

    // If array:
    std::vector<std::pair<long/*beginning index*/, std::size_t/*number of elements*/>> dimensions;

    TypeInfo(TypeType t = TypeType::Integer) : Type(t) {}

    bool is_string() const { return Type == TypeType::String; }
    bool is_num() const    { return Type != TypeType::String
                                 && Type != TypeType::UserType; }
    bool is_array() const  { return arraytype != NotArray; }
    bool is_integer() const { return is_num() && !is_array()
                                  && Type != TypeType::Single
                                  && Type != TypeType::Double; }
    bool is_unsigned() const { return Type == TypeType::UnsignedByte
                                   || Type == TypeType::UnsignedLong
                                   || Type == TypeType::UnsignedBit
                                   || Type == TypeType::UnsignedInteger; }

    bool operator==(const TypeInfo& b) const;
    bool operator!=(const TypeInfo& b) const { return !operator==(b); }

    const std::string Dump() const;
};

struct UserType: public std::vector<std::pair<std::string, TypeP>>
{
    std::unordered_map<std::string, unsigned> Names;

    bool DeclareMember(const std::string& name, TypeP p);
    TypeP GetMember(const std::string& name) const;

    bool operator==(const UserType& b)const;
};


void SetDefVarType(const std::string& list_of_chars, const std::string& suffix);

bool  HasTypeSuffix(const std::string& varname);
std::string CalcTypeSuffix(const std::string& varname, bool force);
std::string StripTypeSuffix(const std::string& varname);
TypeInfo TypeSuffixIntoType(const std::string& type);
std::string TypeIntoTypeSuffix(const TypeInfo& type);

// Creates a type template for the use by DIM and Implicit definitions.
// Uses the type suffix to initialize the "Type" field. Leaves other fields at default.
TypeP CreateTypeTemplate(const std::string& name);

#endif
